import numpy as np
from astropy import units as u
from ctapipe.instrument import CameraGeometry
from ctapipe.instrument.camera import _find_neighbor_pixels, \
    _get_min_pixel_seperation
from numpy import median
import pytest


def test_load_by_name():

    cams = CameraGeometry.get_known_camera_names()
    assert len(cams) > 4
    assert 'FlashCam' in cams
    assert 'NectarCam' in cams
    

    for cam in cams:
        geom = CameraGeometry.from_name(cam)



def test_make_rectangular_camera_geometry():
    geom = CameraGeometry.make_rectangular()
    assert(geom.pix_x.shape == geom.pix_y.shape)


def test_load_hess_camera():
    geom = CameraGeometry.from_name("LSTCam")
    assert len(geom.pix_x) == 1855


def test_guess_camera():
    px = np.linspace(-10, 10, 11328) * u.m
    py = np.linspace(-10, 10, 11328) * u.m
    geom = CameraGeometry.guess(px, py,0 * u.m)
    assert geom.pix_type.startswith('rect')


def test_get_min_pixel_seperation():
    x, y = np.meshgrid(np.linspace(-5, 5, 5), np.linspace(-5, 5, 5))
    pixsep = _get_min_pixel_seperation(x.ravel(), y.ravel())
    assert(pixsep == 2.5)


def test_find_neighbor_pixels():
    x, y = np.meshgrid(np.linspace(-5, 5, 5), np.linspace(-5, 5, 5))
    neigh = _find_neighbor_pixels(x.ravel(), y.ravel(), rad=3.1)
    assert(set(neigh[11]) == set([16, 6, 10, 12]))

def test_neighbor_pixels():
    hexgeom = CameraGeometry.from_name("LSTCam")
    recgeom = CameraGeometry.make_rectangular()

    # most pixels should have 4 neighbors for rectangular geometry and 6 for
    # hexagonal
    assert int(median(recgeom.neighbor_matrix.sum(axis=1))) == 4
    assert int(median(hexgeom.neighbor_matrix.sum(axis=1))) == 6

def test_to_and_from_table():
    geom = CameraGeometry.from_name("LSTCam")
    tab = geom.to_table()
    geom2 = geom.from_table(tab)

    assert geom.cam_id == geom2.cam_id
    assert (geom.pix_x == geom2.pix_x).all()
    assert (geom.pix_y == geom2.pix_y).all()
    assert (geom.pix_area == geom2.pix_area).all()
    assert geom.pix_type == geom2.pix_type


def test_write_read(tmpdir):

    filename = str(tmpdir.join('testcamera.fits.gz'))

    geom = CameraGeometry.from_name("LSTCam")
    geom.to_table().write(filename, overwrite=True)
    geom2 = geom.from_table(filename)

    assert geom.cam_id == geom2.cam_id
    assert (geom.pix_x == geom2.pix_x).all()
    assert (geom.pix_y == geom2.pix_y).all()
    assert (geom.pix_area == geom2.pix_area).all()
    assert geom.pix_type == geom2.pix_type

def test_known_cameras():
    cams = CameraGeometry.get_known_camera_names()
    assert 'FlashCam' in cams
    assert len(cams) > 3


if __name__ == '__main__':
    pass



from ctapipe.core import Component
from ctapipe.io.containers import ReconstructedShowerContainer

__all__ = ['Reconstructor']


class Reconstructor(Component):
    """This is the base class from which all direction reconstruction
algorithms should inherit from"""

    def predict(self, tels_dict):
        """overwrite this method with your favourite direction reconstruction
        algorithm

        Parameters
        ----------
        tels_dict : dict
            general dictionary containing all triggered telescopes data

        Returns
        -------
        Standard  `RecoShowerGeom` container

        """
        return ReconstructedShowerContainer()




from numpy.testing import assert_almost_equal
from ctapipe.io.hessio import hessio_event_source
from ctapipe.utils import get_dataset
from ctapipe.calib.camera.r1 import CameraR1CalibratorFactory, \
    HessioR1Calibrator


def get_test_event():
    filename = get_dataset('gamma_test.simtel.gz')
    source = hessio_event_source(filename, requested_event=409,
                                 use_event_id=True)
    event = next(source)
    return event


def test_hessio_r1_calibrator():
    telid = 11
    event = get_test_event()
    calibrator = HessioR1Calibrator(None, None)
    calibrator.calibrate(event)
    r1 = event.r1.tel[telid].pe_samples
    assert_almost_equal(r1[0, 0, 0], -0.091, 3)


def test_check_r0_exists():
    telid = 11
    event = get_test_event()
    calibrator = HessioR1Calibrator(None, None)
    assert(calibrator.check_r0_exists(event, telid) is True)
    event.r0.tel[telid].adc_samples = None
    assert(calibrator.check_r0_exists(event, telid) is False)


def test_factory():
    factory = CameraR1CalibratorFactory(None, None)
    cls = factory.get_class()
    calibrator = cls(None, None)

    telid = 11
    event = get_test_event()
    calibrator.calibrate(event)
    r1 = event.r1.tel[telid].pe_samples
    assert_almost_equal(r1[0, 0, 0], -0.091, 3)



"""
Classes and functions related to telescope Optics
"""

import logging

logger = logging.getLogger(__name__)

_FOCLEN_TO_TEL_INFO = {
    # foclen: tel_type, tel_subtype, mirror_type
    28.0: ('LST', '', 'DC'),
    16.0: ('MST', '', 'DC'),
    2.28: ('SST', 'GCT', 'SC'),
    2.15: ('SST', 'ASTRI', 'SC'),
    5.6: ('SST', '1M', 'SC'),
    5.58: ('MST', 'SCT', 'SC'),
    5.59: ('MST', 'SCT', 'SC')

}


class OpticsDescription:
    """
    Describes the optics of a Cherenkov Telescope mirror

    The string representation of an `OpticsDescription` will be a combination
    of the telescope-type and sub-type as follows: "type-subtype". You can
    also get each individually.

    The `OpticsDescription.guess()` constructor can be used to fill in info
    from metadata, e.g. for Monte-Carlo files.

    Parameters
    ----------
    mirror_type: str
        'SC' or 'DC'
    tel_type: str
        'SST', 'MST','LST'
    tel_subtype: str
        subtype of telescope, e.g. '1M' or 'ASTRI'
    effective_focal_length: Quantity(float)
        effective focal-length of telescope, independent of which type of
        optics (as in the Monte-Carlo)
    """

    def __init__(self, mirror_type, tel_type,
                 tel_subtype, effective_focal_length,
                 mirror_area=None, num_mirror_tiles=None):

        if tel_type not in ['LST', 'MST', 'SST']:
            raise ValueError("Unknown tel_type %s", tel_type)

        self.mirror_type = mirror_type
        self.tel_type = tel_type
        self.tel_subtype = tel_subtype
        self.effective_focal_length = effective_focal_length
        self.mirror_area = mirror_area
        self.num_mirror_tiles = num_mirror_tiles

    @classmethod
    def guess(cls, effective_focal_length):
        """
        Construct an OpticsDescription by guessing from metadata (e.g. when
        using a simulation where the exact type is not known)

        Parameters
        ----------
        effective_focal_length: Quantity('m')
            effective optical focal-length in meters

        """

        tel_type, tel_subtype, mir_type = \
            telescope_info_from_metadata(effective_focal_length)

        if tel_type == "unknown":
            logger.warning(("No OpticsDescription found for focal-length "
                            "%s, setting to 'unknown'"),
                           effective_focal_length)

        return cls(mirror_type=mir_type,
                   tel_type=tel_type,
                   tel_subtype=tel_subtype,
                   effective_focal_length=effective_focal_length)

    @property
    def identifier(self):
        """ returns a tuple of (tel_type, tel_subtype).  Use str(optics) to
        get a text-based identifier."""
        return (self.tel_type, self.tel_subtype)

    def __repr__(self):
        return "{}(tel_type='{}', tel_subtype='{}')".format(
            str(self.__class__.__name__), self.tel_type, self.tel_subtype)

    def __str__(self):
        if self.tel_subtype != '':
            return "{}-{}".format(self.tel_type, self.tel_subtype)
        else:
            return self.tel_type


def telescope_info_from_metadata(focal_length):
    """
    helper func to return telescope and mirror info based on metadata

    Parameters
    ----------
    focal_length: float
        effective focal length

    Returns
    -------
    str,str,str:
        tel_type ('LST', 'MST' or 'SST'),
        tel_subtype (model),
        mirror_type ('SC' or 'DC')

    Raises:
    -------
    KeyError:
       if unable to find optics type
    """
    return _FOCLEN_TO_TEL_INFO[round(focal_length.to('m').value, 2)]



from ctapipe.calib.camera.dl1 import calibrate_event
from ctapipe.core import Component
from traitlets import Unicode
from traitlets import Int
from traitlets import Float
from traitlets import List

class CalibrationStep(Component):
    """CalibrationStep` class represents a Stage for pipeline.
        it executes ctapipe.calib.camera.calibrators.calibrate_event
        it return calibrated_event and geom_dict
    """
    integrator = Unicode('nb_peak_integration',
     help='integration scheme to be used to extract the charge').tag(
        config=True)
    integration_window = List([7, 3],help='Set integration window width \
        and offset (to before the peak) respectively' ).tag(
        config=True)
    integration_sigamp = List([2, 4], help='Amplitude in ADC counts above \
        pedestal at which a signal is considered as significant, and used \
        for peak finding. (separate for high gain/low gain)').tag(
        config=True)
    integration_clip_amp = Int(default_value=None,allow_none=True, help = 'Amplitude in p.e. above which \
        the signal is clipped').tag(config=True)
    integration_lwt = Int(0, help='Weight of the local pixel (0: peak from \
        neighbours only, 1: local pixel counts as much as any neighbour').tag(
        config=True)
    integration_calib_scale = Float(0.92,help='Used for conversion from ADC to \
        pe. Identical to global variable CALIB_SCALE in reconstruct.c in \
        hessioxxx software package.').tag(config=True)


    def init(self):
        self.log.debug("--- CalibrationStep init ---")
        self.parameters = dict()
        self.parameters['integrator'] = self.integrator#'nb_peak_integration'
        self.parameters['integration_window'] = self.integration_window
        self.parameters['integration_shift'] =  self.integration_window[1]
        self.parameters['integration_sigamp'] = self.integration_sigamp
        self.parameters['integration_lwt'] = self.integration_lwt
        self.parameters['integration_calib_scale'] = self.integration_calib_scale

        if self.integration_clip_amp != None:
                self.parameters['clip_amp'] = self.integration_clip_amp

        for key, value in sorted(self.parameters.items()):
            self.log.info("[{}] {}".format(key, value))
        return True

    def run(self, event):
        if event != None:
            self.log.debug("--- CalibrationStep RUN --- event {}".format(event.dl0.event_id))
            geom_dict = {}
            calibrated_event = calibrate_event(event,self.parameters,geom_dict)
            #for tel_id in calibrated_event.r0.tels_with_data:
            #    signals = calibrated_event.dl1.tel[tel_id].calibrated_image
            #    cmaxmin = (max(signals) - min(signals))
            self.log.debug("--- CalibrationStep STOP ---")
            return ([calibrated_event,geom_dict])

    def finish(self):
        self.log.debug("--- CalibrationStep finish ---")
        pass



from copy import deepcopy
from os import remove

import pytest
from astropy.io import fits

from ctapipe.io.hessio import hessio_event_source
from ctapipe.io.serializer import Serializer
from ctapipe.io.sources import PickleSource
from ctapipe.utils import get_dataset


def compare(read_container, source_container):
    # test if 4th adc value of telescope 17 HI_GAIN are equals
    return (read_container.r0.tel[17].adc_samples[0][2][4] ==
            source_container.r0.tel[17].adc_samples[0][2][4])


def generate_input_containers():
    # Get event from hessio file, append them into input_containers
    input_filename = get_dataset("gamma_test.simtel.gz")
    gen = hessio_event_source(input_filename, max_events=3)
    input_containers = []
    for event in gen:
        input_containers.append(deepcopy(event))
    return input_containers

# Setup
input_containers = generate_input_containers()

@pytest.fixture(scope='session')
def binary_filename(tmpdir_factory):
    return str(tmpdir_factory.mktemp('data')
               .join('pickle_data.pickle.gz'))

@pytest.fixture(scope='session')
def fits_file_name(tmpdir_factory):
    return str(tmpdir_factory.mktemp('data').join('output.fits'))



def test_pickle_serializer(binary_filename):
    serial = Serializer(filename=binary_filename, format='pickle', mode='w')
    # append all input file events in input_containers list and pickle serializer
    for event in input_containers:
        serial.add_container(event)
    serial.close()

    # read Containers from pickle serializer
    reader = PickleSource(filename=binary_filename)
    # file read_containers from serializer generator
    read_containers = []
    for container in reader:
        read_containers.append(container)
    # test if number of read Container correspond to input
    assert len(read_containers) is len(input_containers)
    # test if 4th adc value of telescope 17 HI_GAIN are equals
    assert compare(input_containers[2], read_containers[2])
    reader.close()
    remove(binary_filename)


# Test pickle reader/writer with statement
def test_pickle_with_statement(binary_filename):
    with Serializer(filename=binary_filename, format='pickle', mode='w') as \
            containers_writer:
        for container in input_containers:
            containers_writer.add_container(container)
        containers_writer.close()

    read_containers = []
    with PickleSource(filename=binary_filename) as reader:
        for container in reader:
            read_containers.append(container)
    # test if number of read Container correspond to input
    assert len(read_containers) is len(input_containers)
    # test if 4th adc value of telescope 17 HI_GAIN are equals
    assert compare(input_containers[2], read_containers[2])
    remove(binary_filename)


# Test pickle reader iterator
def test_pickle_iterator(binary_filename):
    serial = Serializer(filename=binary_filename, format='pickle',
                        mode='w')
    # append all events in input_containers list and pickle serializer
    for event in input_containers:
        serial.add_container(event)
    serial.close()

    read_containers = []
    reader = PickleSource(filename=binary_filename)
    for container in reader:
        read_containers.append(container)
    # test if number of read Container correspond to input
    assert len(read_containers) is len(input_containers)
    # test if 4th adc value of telescope 17 HI_GAIN are equals
    assert compare(input_containers[2], read_containers[2])
    reader.close()
    remove(binary_filename)





def test_fits_dl0(fits_file_name):
    serial = Serializer(filename=fits_file_name, format='fits', mode='w')
    for container in input_containers:
        serial.add_container(container.dl0)
    serial.close()
    hdu = fits.open(fits_file_name)[1]
    assert hdu.data["event_id"][0] == 408
    assert hdu.data["event_id"][1] == 409
    assert hdu.data["event_id"][2] == 803
    assert hdu.data["run_id"][2] == 31964
    remove(fits_file_name)


def test_exclusive_mode(fits_file_name):
    serial = Serializer(filename=fits_file_name, format='fits', mode='w')
    for container in input_containers:
        serial.add_container(container.dl0)
    serial.close()
    # Try to write to fits_file_name in exclusive mode
    with pytest.raises(OSError):
        serial = Serializer(filename=fits_file_name, format='fits', mode='x')
        serial.add_container(input_containers[2].dl0)
        serial.close()
    remove(fits_file_name)

"""
def test_fits_dl1():
    input_test_file = get_datasets_path('example_container.pickle.gz')
    with gzip_open(input_test_file, 'rb') as f:
        data = load(f)
    t38 = data[0].dl1.tel[38]
    serial = Serializer('output.fits', 'fits', overwrite=True)
    serial.add_container(t38)
    serial.write()
    # t11_1 = data[1].dl1.tel[11]
    # S_cal.write(t11_1) # This will not work because shape of data is different from tel to tel.
"""


def test_fits_context_manager(fits_file_name):
    with Serializer(filename=fits_file_name, format='fits', mode='w') as writer:
        for container in input_containers:
            writer.add_container(container.dl0)

    hdulist = fits.open(fits_file_name)
    assert hdulist[1].data["event_id"][0] == 408
    remove(fits_file_name)


# TODO test FITSSource class


import astropy.units as u
from astropy.table import Table
from scipy.optimize import minimize
import scipy.integrate as integrate
import numpy as np


__all__ = ["SensitivityPointSource"]


def crab_source_rate(energy):
    '''
    function for a pseudo-Crab point-source rate:
        dN/dE = 3e-7  * (E/TeV)**-2.5 / (TeV * m² * s)
    (watch out: unbroken power law... not really true)
    norm and spectral index reverse engineered from HESS plot...

    Parameters
    ----------
    energy : astropy quantity
        energy for which the rate is desired

    Returns
    -------
    flux : astropy quantity
        differential flux at E

    '''
    return 3e-7 * (energy / u.TeV)**-2.5 / (u.TeV * u.m**2 * u.s)


def cr_background_rate(energy):
    '''
    function for the Cosmic Ray background rate:
        dN/dE = 0.215 * (E/TeV)**-8./3 / (TeV * m² * s * sr)
    (simple power law, no knee/ankle)
    norm and spectral index reverse engineered from "random" CR plot...

    Parameters
    ----------
    energy : astropy quantity
        energy for which the rate is desired

    Returns
    -------
    flux : astropy quantity
        differential flux at E

    '''
    return 100 * 0.1**(8. / 3) * (energy / u.TeV)**(-8. / 3) / \
        (u.TeV * u.m**2 * u.s * u.sr)


def e_minus_2(energy, unit=u.TeV):
    '''
    boring, old, unnormalised E^-2 spectrum

    Parameters
    ----------
    energy : astropy quantity
        energy for which the rate is desired

    Returns
    -------
    flux : astropy quantity
        differential flux at E

    '''
    return (energy / unit)**(-2) / (unit * u.s * u.m**2)


def make_mock_event_rate(spectrum, bin_edges, log_e=False, norm=None):
    """
    Creates a histogram with a given binning and fills it according to a spectral function

    Parameters
    ----------
    spectrum : function object
        function of the differential spectrum that shall be sampled into the histogram
        ought to take the energy as an astropy quantity as sole argument
    bin_edges : numpy array, optional (default: None)
        bin edges of the histogram that is to be filled
    log_e : bool, optional (default: False)
        tell if the values in `bin_edges` are given in logarithm
    norm : float, optional (default: None)
        normalisation factor for the histogram that's being filled
        sum of all elements in the array will be equal to `norm`

    Returns
    -------
    rates : numpy array
        histogram of the (non-energy-differential) event rate of the proposed spectrum
    """

    def spectrum_value(e):
        """
        `scipy.integrate` does not like units during integration. use this as a quick fix
        """
        return spectrum(e).value

    rates = []
    if log_e:
        bin_edges = 10**bin_edges
    for l_edge, h_edge in zip(bin_edges[:-1], bin_edges[1:]):
        bin_events = integrate.quad(
            spectrum_value, l_edge.value, h_edge.value)[0]
        rates.append(bin_events)

    # units have been strip for the integration. the unit of the result is the unit of the
    # function: spectrum(e) times the unit of the integrant: e -- for the latter use the
    # first entry in `bin_edges`
    rates = np.array(rates) * spectrum(bin_edges[0]).unit * bin_edges[0].unit

    # if `norm` is given renormalise the sum of the `rates`-bins to this value
    if norm:
        rates *= norm / np.sum(rates)

    return rates


def sigma_lima(n_on, n_off, alpha=0.2):
    """
    Compute the significance according to Eq. (17) of Li & Ma (1983).

    Parameters
    ----------
    n_on : integer/float
        Number of on counts
    n_off : integer/float
        Number of off counts
    alpha : float, optional (default: 0.2)
        Ratio of on-to-off exposure

    Returns
    -------
    sigma : float
        the significance of the given off and on counts
    """

    alpha1 = alpha + 1.0
    n_sum = n_on + n_off
    arg1 = n_on / n_sum
    arg2 = n_off / n_sum
    term1 = n_on
    term2 = n_off
    if n_on > 0:
        term1 *= np.log((alpha1 / alpha) * arg1)
    if n_off > 0:
        term2 *= np.log(alpha1 * arg2)
    sigma = np.sqrt(2.0 * (term1 + term2))

    return sigma


def diff_to_x_sigma(scale, n, alpha, x=5):
    """
    calculates the significance according to `sigma_lima` and returns the squared
    difference to `x`. To be used in a minimiser that determines the necessary source
    intensity for a detection of given significance of `x` sigma
    The square here is only to have a continuously differentiable function with a smooth
    turning point at the minimum -- in contrast to an absolute-function that makes a
    sharp turn at the minimum.

    Parameters
    ----------
    scale : python list with a single float
        this is the variable in the minimisation procedure
        it scales the number of gamma events
    n : shape (2) list
        the signal count in the on- (index 0)
        and the background count int the off-region (index 1)
        the events in the on-region are to be scaled by `scale[0]`
        the background rate in the on-region is estimated as `alpha` times off-count
    alpha : float
        the ratio of the on and off areas
    x : float, optional (default: 5)
        target significance in multiples of "sigma"

    Returns
    -------
    (sigma - x)**2 : float
        squared difference of the significance to `x`
        minimise this function for the number of gamma events needed for your desired
        significance of `x` sigma
    """

    n_on = n[1] * alpha + n[0] * scale[0]
    n_off = n[1]
    sigma = sigma_lima(n_on, n_off, alpha)
    return (sigma - x)**2


class SensitivityPointSource:
    """
    class to calculate effective areas and the sensitivity to a known point-source
    TODO extend sensitvity module
        • add extended source?
        • add pseudo experiment for low exposure times?
        ✔ add methods for time stamp generation and energy sampling of events

    Parameters
    ----------
    reco_energies: dictionary of quantified numpy arrays, optional
        lists of reconstructed of the selected events for each channel
    mc_energies: dictionary of quantified numpy arrays, optional
        lists of simulated energies of the selected events for each channel
    energy_bin_edges : dictionary of numpy arrays, optional
        lists of the bin edges for the various histograms for each channel;
        assumes binning in log10(energy)
        you should provide this at some point if not done here
    energy_unit : astropy quantity, optional (default: u.TeV)
        your favourite energy unit
    flux_unit : astropy quantity, optional (default: 1 / (u.TeV * u.m**2 * u.s))
        your favourite differential flux unit
    """

    def __init__(self, reco_energies=None, mc_energies=None, energy_bin_edges=None,
                 energy_unit=u.TeV, flux_unit=1 / (u.TeV * u.m**2 * u.s)):

        self.reco_energies = reco_energies
        self.mc_energies = mc_energies
        self.energy_bin_edges = energy_bin_edges

        try:
            self.class_list = reco_energies.keys()
        except AttributeError:
            self.class_list = mc_energies.keys()

        self.energy_unit = energy_unit
        self.flux_unit = flux_unit

    def get_effective_areas(self, generator_areas,
                            n_simulated_events=None,
                            generator_spectra=None,
                            generator_energy_hists=None
                            ):
        """
        calculates the effective areas for the provided channels and stores them in the
        class instance

        Parameters
        ----------
        generator_areas : astropy quantities
            the area for each channel within which the shower impact position was
            generated
        n_simulated_events : dictionary of integers, optional (defaults: None)
            number of events used in the MC simulation for each channel
        generator_spectra : dictionary of functors, optional (default: None)
            function object for the differential generator flux of each channel
        generator_energy_hists : numpy arrays, optional (default: None)
            energy histogram of the generated events for each channel binned according to
            `.energy_bin_edges`

        Returns
        -------
        eff_area_gam, eff_area_pro : numpy arrays
            histograms of the effective areas of gammas and protons binned according to
            `.bin_edges_gam` and `.bin_edges_pro`

        Notes
        -----
        either give the histogram of the energy distributions at MC generator level with
        `generator_energy_hists` or create them on the fly with `n_simulated_events` and
        `spectra`
        """

        if (n_simulated_events is not None and generator_spectra is not None) == \
                (generator_energy_hists):
            raise ValueError("use either (n_simulated_events and generator"
                             "_spectra) or generator_energy_hists to set the MC "
                             "generated energy spectrum -- not both")

        if not generator_energy_hists:
            generator_energy_hists = {}
            # generate the histograms for the energy distributions of the Monte Carlo
            # generated events given the generator spectra and the number of generated
            # events
            for cl in self.class_list:
                generator_energy_hists[cl] = make_mock_event_rate(
                    generator_spectra[cl], norm=n_simulated_events[cl],
                    bin_edges=self.energy_bin_edges[cl], log_e=False)
        self.generator_energy_hists = generator_energy_hists

        # an energy-binned histogram of the effective areas
        # binning according to .energy_bin_edges[cl]
        self.effective_areas = {}

        # an energy-binned histogram of the selected events
        # binning according to .energy_bin_edges[cl]
        self.selected_events = {}

        # generate the histograms for the energy distributions of the selected
        # events
        for cl in self.class_list:
            self.selected_events[cl] = np.histogram(self.mc_energies[cl],
                                                    bins=self.energy_bin_edges[cl])[0]

            # the effective areas are the selection efficiencies per energy bin multiplied
            # by the area in which the Monte Carlo events have been generated
            # in
            efficiency = self.selected_events[
                cl] / self.generator_energy_hists[cl]
            self.effective_areas[cl] = efficiency * generator_areas[cl]

        return self.effective_areas

    def generate_event_weights_histogram(self, spectra=None, extensions=None,
                                         observation_time=50 * u.h):
        """
        given a source rate and the effective area, calculates the number of expected
        events within a given observation time
        then, produces weights to scale the selected events according to source/background
        spectrum and observation time

        Parameters
        ----------
        spectra : dictionary of functors, optional (default: 'g': `crab_source_rate`,
                                                             'p': `cr_background_rate`)
            functions for the differential source and background rates
        extensions : dictionary of astropy quantities, optional (default: {'p': 6*u.deg})
            opening angle of the view-cone the events have been generated in, if any
            don't set the key if a channel was generated as a point-source
            note: if you use an extension, the flux needs to accomodate that as well
        observation_time : astropy quantity, optional (default: 50*u.h)
            length of the assumed exposure

        Returns
        -------
        weights : dictionary of numpy arrays
            weights for the selected events so that they are scaled to the number of
            expected events in `exp_events_per_energy_bin` for every energy bin
        """

        # setting defaults because mutable
        spectra = spectra or {'g': crab_source_rate, 'p': cr_background_rate},
        extensions = extensions or {'p': 6 * u.deg}

        # an energy-binned histogram of the number of events expected from the given flux
        # binning according to .energy_bin_edges[cl]
        self.exp_events_per_energy_bin = {}
        for cl in self.class_list:
            event_rate = make_mock_event_rate(spectra[cl], log_e=False,
                                              bin_edges=self.energy_bin_edges[cl])
            if cl in extensions:
                omega = 2 * np.pi * (1 - np.cos(extensions[cl])) * u.rad**2
                event_rate *= omega

            self.exp_events_per_energy_bin[cl] = (event_rate * observation_time *
                                                  self.effective_areas[cl]).si

        self.event_weights = {}
        for cl in self.class_list:
            weights = (self.exp_events_per_energy_bin[
                       cl] / self.selected_events[cl]).si
            self.event_weights[cl] = weights[np.clip(
                                np.digitize(self.mc_energies[cl],
                                            self.energy_bin_edges[cl]) - 1,
                                0, len(self.energy_bin_edges[cl]) - 2)]

            self.event_weights[cl] = np.array(self.event_weights[cl])

        return self.event_weights

    def generate_event_weights(self, n_simulated_events, e_min_max, spectra,
                               generator_areas,
                               observation_time=50 * u.h,
                               generator_gamma=None,
                               extensions=None):
        """
        generates a weight for every event

        Parameters
        ----------
        n_simulated_events : dictionary
            total number of simulated events for every channel
        e_min_max : dictionary of tuples
            lower and upper energy limits used in the Monte Carlo event generator
        generator_areas : dictionary
            area within which the shower impact point has been distributed by the
            Monte Carlo event generator
        extensions : dictionary, optional (default: {'p': 6*u.deg})
        """

        # setting defaults because mutable
        generator_gamma = generator_gamma or {"g": 2, "p": 2}
        extensions = extensions or {'p': 6 * u.deg}

        self.event_weights = {}
        self.exp_events_per_energy_bin = {}
        for cl in self.class_list:
            # event weights for a flat energy distribution
            e_w = (self.mc_energies[cl])**generator_gamma[cl] \
                  * (e_min_max[cl][1]**(1 - generator_gamma[cl]) -
                     e_min_max[cl][0]**(1 - generator_gamma[cl])) / \
                    (1 - generator_gamma[cl]) \
                  * generator_areas[cl] \
                  * (1 if (cl not in extensions) else
                     2 * np.pi * (1 - np.cos(extensions[cl])) * u.rad**2) \
                  * observation_time \
                  / n_simulated_events[cl]

            # multiply these flat event weights by the flux to get weights corresponding
            # to the number of expected events from that flux
            # the event weight should be unit-less; the call to `dimensionless_unscaled`
            # is an explicit test of this requirement and resolves any possible mixed
            # units to unity
            # QUESTION use `.si` to resolve mixed units to allow unified weights?
            # e.g. `observation_time` could be set to `1` -- without unit -- to get
            # `event_weights` in units of 1/time and be later multiplied by different
            # observation times
            self.event_weights[cl] = \
                (e_w * spectra[cl](self.mc_energies[cl])
                 ).to(u.dimensionless_unscaled)

            # now, for the fun of it, make an energy-binned histogram of the
            # events
            self.exp_events_per_energy_bin[cl], _ = \
                np.histogram(self.mc_energies[cl],
                             bins=self.energy_bin_edges[cl],
                             weights=self.event_weights[cl])

        return self.event_weights

    def get_sensitivity(self, sensitivity_energy_bin_edges,
                        alpha, signal_list=("g"), mode="MC",
                        sensitivity_source_flux=crab_source_rate,
                        min_n=10, max_background_ratio=.05):
        """
        finally calculates the sensitivity to a point-source
        TODO still need to implement statistical error on the sensitivity

        Parameters
        ----------
        alpha : float
            area-ratio of the on- over the off-region
        sensitivity_energy_bin_edges : numpy array
            array of the bin edges for the sensitivity calculation
        mode : string ["MC", "Data"] (default: "MC")
            interprete the signal/not-signal channels in all the dictionaries as
            gamma/background ("MC") or as on-region/off-region ("Data")
            - if "MC":
                the signal channel is taken as the part comming from the source and
                the background channels multiplied by `alpha` is used as the background
                part in the on-region; the background channels themselves are taken as
                coming from the off-regions
            - if "Data":
                the signal channel is taken as the counts reconstructed in the on-region
                the counts from the background channels multiplied by `alpha` are taken as
                the background estimate for the on-region
        min_n : integer, optional (default: 10)
            minimum number of events per energy bin -- if the number is smaller, scale up
            all events to sum up to this
        max_background_ratio : float, optional (default: 0.05)
            maximal background contamination per bin -- if fraction of protons in a bin
            is larger than this, scale up the gammas events accordingly
        sensitivity_source_flux : callable, optional (default: `crab_source_rate`)
            function of the flux the sensitivity is calculated with

        Returns
        -------
        sensitivities : astropy.table.Table
            the sensitivity for every energy bin of `sensitivity_energy_bin_edges`
        """

        # sensitivities go in here
        sensitivities = Table(
            names=("Energy", "Sensitivity", "Sensitivity_base"))
        try:
            sensitivities["Energy"].unit = sensitivity_energy_bin_edges.unit
        except AttributeError:
            sensitivities["Energy"].unit = self.energy_unit
        sensitivities["Sensitivity"].unit = self.flux_unit
        sensitivities["Sensitivity_base"].unit = self.flux_unit

        if hasattr(self, "event_weights"):
            # in case we do have event weights, we sum them within the energy bin
            # count also the square of the weights for the error estimator
            count_events = lambda mask: np.sum(self.event_weights[cl][mask])
            count_square = lambda mask: np.sum(self.event_weights[cl][mask]**2)
        else:
            # otherwise we simply check the length of the masked energy array
            # since the weights are 1 here, `count_square` is the same as
            # `count_events`
            count_events = lambda mask: len(self.reco_energies[cl][mask])
            count_square = count_events

        # loop over all energy bins
        # the bins are spaced logarithmically: use the geometric mean as the bin-centre,
        # so when plotted logarithmically, they appear at the middle between
        # the bin-edges
        for elow, ehigh, emid in zip(sensitivity_energy_bin_edges[:-1],
                                     sensitivity_energy_bin_edges[1:],
                                     np.sqrt(sensitivity_energy_bin_edges[:-1] *
                                             sensitivity_energy_bin_edges[1:])):

            N_events = np.zeros(2)  # [on-signal, off-background]
            variance = np.zeros(2)  # [on-signal, off-background]

            # count the (weights of the) events in the on and off regions for this
            # energy bin
            for cl in self.class_list:
                # single out the events in this energy bin
                e_mask = (self.reco_energies[cl] > elow) & \
                         (self.reco_energies[cl] < ehigh)

                # count the events as the sum of their weights within this
                # energy bin
                if cl in signal_list:
                    N_events[0] += count_events(e_mask)
                    variance[0] += count_square(e_mask)
                else:
                    N_events[1] += count_events(e_mask)
                    variance[1] += count_square(e_mask)

            if mode.lower() == "data":
                # the background estimate for the on-region is `alpha` times the
                # background in the off-region
                # if running on data, the signal estimate for the on-region is the counts
                # in the on-region minus the background estimate for the on-region
                N_events[0] -= N_events[1] * alpha
                variance[0] -= variance[1] * alpha

            # If we have no counts in the on-region, there is no sensitivity.
            # If on data the background estimate from the off-region is larger than the
            # counts in the on-region, `sigma_lima` will break! Skip those
            # cases, too.
            if N_events[0] <= 0:
                continue

            # find the scaling factor for the gamma events that gives a 5 sigma discovery
            # in this energy bin
            scale = minimize(diff_to_x_sigma, [1e-3],
                             args=(N_events, alpha),
                             method='L-BFGS-B', bounds=[(1e-4, None)],
                             options={'disp': False}
                             ).x[0]

            scale_base = scale

            # scale up the gamma events by this factor
            N_events[0] *= scale

            # check if there are sufficient events in this energy bin
            scale *= check_min_n(N_events, min_n=min_n, alpha=alpha)

            # check if the relative amount of protons in this bin is
            # sufficiently small
            scale *= check_background_contamination(
                N_events, alpha=alpha, max_background_ratio=max_background_ratio)

            # get the flux at the bin centre
            flux = sensitivity_source_flux(emid).to(self.flux_unit)

            # and scale it up by the determined factor
            sensitivity = flux * scale
            sensitivity_base = flux * scale_base

            # store results in table
            sensitivities.add_row([emid, sensitivity, sensitivity_base])

        return sensitivities

    def calculate_sensitivities(self,
                                # arguments for `get_effective_areas`
                                generator_areas,
                                n_simulated_events=None,
                                generator_spectra=None,
                                generator_energy_hists=None,

                                # arguments for `generate_event_weights`
                                extensions=None,
                                observation_time=50 * u.h,
                                generator_gamma=None,
                                e_min_max=None,
                                spectra=None,

                                # arguments for `get_sensitivity`
                                sensitivity_energy_bin_edges=np.logspace(-1, 3, 17)*u.TeV,
                                alpha=1, min_n=10, max_background_ratio=.05,
                                ):
        """
        wrapper that calls all functions to calculate the point-source sensitivity

        Parameters
        ----------
        cf. corresponding functions:
        `.get_effective_areas`
        `.generate_event_weights`
        `.get_sensitivity`


        Returns
        -------
        sensitivities : `astropy.table.Table`
            the sensitivity for every energy bin of `.bin_edges_gam`
        """

        generator_energy_hists = generator_energy_hists or {}
        extensions = extensions or {'p': 6 * u.deg}



        # this is not needed for the sensitivity, but effective areas are nice to have so,
        # do it anyway here
        self.get_effective_areas(n_simulated_events=n_simulated_events,
                                 generator_spectra=generator_spectra,
                                 generator_energy_hists=generator_energy_hists,
                                 generator_areas=generator_areas)

        self.generate_event_weights(n_simulated_events=n_simulated_events,
                                    observation_time=observation_time,
                                    extensions=extensions,
                                    e_min_max=e_min_max,
                                    spectra=spectra,
                                    generator_gamma=generator_gamma,
                                    generator_areas=generator_areas)

        return self.get_sensitivity(
            alpha=alpha, min_n=min_n,
            max_background_ratio=max_background_ratio,
            sensitivity_energy_bin_edges=sensitivity_energy_bin_edges)

    @staticmethod
    def generate_toy_timestamps(light_curves, time_window):
        """
        randomly draw time stamps within a time window
        either uniformly distributed or by following the distribution of a light curve

        Parameters
        ----------
        light_curve : dictionary of numpy arrays or floats
            light curve you want to sample time stamps from
            sum of the elements (cast to `int`) will be taken as number of draws
            if value is float and not an array, assume take than uniform draws
        time_window : tuple of floats
            min and max of the drawn time stamps

        Returns
        -------
        time_stamps : dictionary of numpy arrays
            lists of randomly drawn time stamps
        """

        time_stamps = {}
        for cl, f in light_curves.items():
            if hasattr(f, "__iter__"):
                # if `f` is any kind of iterable container, create cumulative
                # distribution, draw  randomly from it, and sample time stamps within the
                # `time_window`  accordingly
                cum_sum = np.cumsum(f)
                random_draws = np.random.uniform(
                    0, cum_sum[-1], int(cum_sum[-1]))
                indices = np.digitize(random_draws, cum_sum)
                time_stamps[cl] = time_window[0] + indices * (time_window[1] -
                                                              time_window[0]) / len(f)
            else:
                # if `f` is just a number, draw uniformly that many events within the
                # `time_window`
                time_stamps[cl] = np.random.uniform(*time_window, f)

        return time_stamps

    @staticmethod
    def draw_events_from_flux_histogram(mc_energies, spectrum_hists, energy_bin_edges):

        drawn_indices = {}
        for cl, h in spectrum_hists.items():
            cum_sum = np.insert(np.cumsum(h), 0, 0)

            # rounding errors can make the sum slightly smaller than intended
            # add a small epsilon to push above that number
            random_draws = np.random.uniform(0, cum_sum[-1], int(cum_sum[-1] + 1e-5))

            cumsum_indices = np.digitize(random_draws, cum_sum)

            cum_sums_up = cum_sum[cumsum_indices]
            cum_sums_lo = cum_sum[cumsum_indices - 1]

            energies_up = energy_bin_edges[cl][cumsum_indices]
            energies_lo = energy_bin_edges[cl][cumsum_indices - 1]

            drawn_energies = energies_lo + (random_draws - cum_sums_lo) * \
                (energies_up - energies_lo) / (cum_sums_up - cum_sums_lo)

            # here be numpy magic:
            # `drawn_energies[:, np.newaxis]` enables the broadcast to `mc_energies`
            #
            # `np.abs(mc_energies[cl] - drawn_energies[:, np.newaxis])` is a matrix
            # containing the absolute difference between every drawn energy with every MC
            # energy
            #
            # `np.argmin(X, axis=1)` returns the indices along the `mc_energies` axis of
            # the elements with the smallest difference, i.e. these are the MC events with
            # the energy closest to the randomly drawn energies
            drawn_indices[cl] = np.argmin(np.abs(mc_energies[cl] -
                                                 drawn_energies[:, np.newaxis]),
                                          axis=1)
        return drawn_indices

    @staticmethod
    def draw_events_from_flux_weight(mc_energies, energy_weights, N_draws):

        drawn_indices = {}
        for cl, weights in energy_weights.items():
            drawn_indices[cl] = np.random.choice(range(len(mc_energies[cl])),
                                                 size=N_draws[cl],
                                                 p=weights)
        return drawn_indices


def check_min_n(n, alpha=1, min_n=10):
    """
    check if there are sufficenly many events in this energy bin and calculates scaling
    parameter if not.
    scales `N_signal` in place.

    Parameters
    ----------
    n = [n_signal, n_backgr] : list
        the signal in the on-region (index 0) and background in the off-region (index 1)
    alpha : float (default: 1)
        ratio of the on- to off-region -- `n[1]` times `alpha` is used as background
        estimate for the on-region
    min_n : integer (default: 10)
        minimum number of desired events; if too low, scale up to this number

    Returns
    -------
    scale_a : float
        factor to scale up gamma events if insuficently many events present
    """

    n_signal, n_backgr = n[0], n[1] * alpha

    if n_signal + n_backgr < min_n:
        scale_a = (min_n - n_backgr) / n_signal
        n[0] *= scale_a
        return scale_a
    else:
        return 1


def check_background_contamination(n, alpha=1, max_background_ratio=.05):
    """
    check if there are sufficenly few proton events in this energy bin and calculates
    scaling parameter if not
    scales `N_signal` in place.

    Parameters
    ----------
    n = [n_signal, n_backgr] : list
        the signal in the on-region (index 0) and background in the off-region (index 1)
    alpha : float (default: 1)
        ratio of the on- to off-region -- `n[1]` times `alpha` is used as background
        estimate for the on-region
    max_background_ratio : float
        maximum allowed proton contamination

    Returns
    -------
    scale_r : float
        factor to scale up gamma events if too high proton contamination
    """

    n_signal, n_backgr = n[0], n[1] * alpha

    n_tot = n_signal + n_backgr
    if n_backgr / n_tot > max_background_ratio:
        scale_r = (1 / max_background_ratio - 1) * n_backgr / n_signal
        n[0] *= scale_r
        return scale_r
    else:
        return 1



import logging


class ColoredFormatter(logging.Formatter):
    """
    Custom logging.Formatter that adds colors in addition to the original
    Application logger functionality from LevelFormatter (in application.py)
    """
    highlevel_limit = logging.WARN
    highlevel_format = " %(levelname)s |"

    def format(self, record):
        black, red, green, yellow, blue, magenta, cyan, white = range(8)
        reset_seq = "\033[0m"
        color_seq = "\033[1;%dm"
        colors = {
            'WARNING': yellow,
            'INFO': green,
            'DEBUG': blue,
            'CRITICAL': yellow,
            'ERROR': red
        }

        levelname = record.levelname
        if levelname in colors:
            levelname_color = color_seq % (30 + colors[levelname]) \
                + levelname + reset_seq
            record.levelname = levelname_color

        if record.levelno >= self.highlevel_limit:
            record.highlevel = self.highlevel_format % record.__dict__
        else:
            record.highlevel = ""

        return super(ColoredFormatter, self).format(record)



from ctapipe.instrument.optics import  OpticsDescription
from astropy import units as u
import pytest

def test_guess_optics():

    od = OpticsDescription.guess(28.0*u.m)

    assert od.tel_type == 'LST'
    assert od.tel_subtype == ''
    assert od.mirror_type == 'DC'

    with pytest.raises(KeyError):
        od2 = OpticsDescription.guess(0*u.m)




from ctapipe.calib.camera.dl1 import integration_correction, \
    CameraDL1Calibrator
from ctapipe.calib.camera.dl0 import CameraDL0Reducer
from ctapipe.calib.camera.r1 import HessioR1Calibrator
from ctapipe.io.hessio import hessio_event_source
from ctapipe.utils import get_dataset
from numpy.testing import assert_allclose


def get_test_event():
    filename = get_dataset('gamma_test.simtel.gz')
    source = hessio_event_source(filename, requested_event=409,
                                 use_event_id=True)
    event = next(source)
    return event


def previous_calibration(event):
    r1 = HessioR1Calibrator(None, None)
    r1.calibrate(event)
    dl0 = CameraDL0Reducer(None, None)
    dl0.reduce(event)


def test_integration_correction():
    event = get_test_event()
    telid = 11
    width = 7
    shift = 3
    n_chan = event.inst.num_channels[telid]
    shape = event.mc.tel[telid].reference_pulse_shape
    step = event.mc.tel[telid].meta['refstep']
    time_slice = event.mc.tel[telid].time_slice
    correction = integration_correction(n_chan, shape, step,
                                        time_slice, width, shift)
    assert_allclose(correction[0], 1.077, 1e-3)


def test_camera_dl1_calibrator():
    event = get_test_event()
    previous_calibration(event)
    telid = 11

    calibrator = CameraDL1Calibrator(None, None)

    correction = calibrator.get_correction(event, telid)
    assert_allclose(correction[0], 1.077, 1e-3)

    calibrator.calibrate(event)
    image = event.dl1.tel[telid].image
    assert_allclose(image[0, 0], -2.216, 1e-3)


def test_check_dl0_exists():
    telid = 11
    event = get_test_event()
    previous_calibration(event)
    calibrator = CameraDL1Calibrator(None, None)
    assert(calibrator.check_dl0_exists(event, telid) is True)
    event.dl0.tel[telid].pe_samples = None
    assert(calibrator.check_dl0_exists(event, telid) is False)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
from scipy import ndimage
from astropy.io import fits
from astropy.wcs import WCS

__all__ = ['Histogram']


class Histogram:
    """An N-D histogram class with FITS image I/O.

    The output FITS file will contain an ImageHDU datacube and
    associated WCS headers to describe the axes of the histogram.
    Thus, the output files should work correctly in any program
    capable of working with FITS datacubes (like SAOImage DS9).

    Internally, it uses `numpy.histogramdd` to generate the histograms.

    All axes are assumed to be linear, with equally spaced bins
    (otherwise they could not be stored in a FITS image HDU)

    Parameters
    ----------
    nbins: array_like(int)
        list of number of bins for each dimension
    ranges: list(tuple)
        list of (min,max) values for each dimension
    name: str
        name of histogram (will be used as FITS extension name when written
        to a file
    axis_names: list(str)
        name of each axis

    Examples
    --------

    >>> hist = Histogram(nbins=(10,10), ranges=[[-1,1], [-1,1]])
    >>> data = np.random.normal(shape=(2*100)) # make 100 random 2D events
    >>> hist.fill(data)


    Get a point in the histogram (can also get multiple values at once by
    passing an array)

    >>> val = hist.get_value([0.1,-0.5])
    >>> vals = hist.get_value([[0.1,-0.5], [0.9,0.9]])

    Get the full data array and do things with it:

    >>> meanx = hist.data.mean(axis=0)

    Write it to a FITS image file:

    >>> hist.to_fits().writeto("output.fits")

    Read it from FITS image file:

    >>> hist2 = Histogram.from_fits("output.fits")

    """

    def __init__(self, nbins=None, ranges=None, name="Histogram",
                 axis_names=None):
        """ Initialize an unfilled histogram (need to call fill()  put data into it)

        see also
        --------
        The `Histogram.from_fits()` constructor
        """

        self.data = np.zeros(nbins)
        self._bin_lower_edges = None
        self._nbins = np.array([nbins]).flatten()
        self._ranges = np.array(ranges, ndmin=2)
        self.value_scale = None
        self.value_zero = None
        self.name = name
        self._ctypes = None
        self.axis_names = axis_names
        self._numsamples = 0

        # sanity check on inputs:

        if self.ndims < 1:
            raise ValueError("No dimensions specified")
        if self.ndims != len(self._ranges):
            raise ValueError("Dimensions of ranges {0} don't match bins {1}"
                             .format(len(self._ranges), self.ndims))

        if self.axis_names is not None:  # ensure the array is size ndims
            self.axis_names = np.array(self.axis_names)
            self.axis_names.resize(self.ndims)
        else:
            self.axis_names = ["axis{}".format(x) for x in range(self.ndims)]

    def __str__(self,):
        return ("Histogram(name='{name}', axes={axnames}, "
                "nbins={nbins}, ranges={ranges})"
                .format(name=self.name, ranges=self._ranges,
                        nbins=self._nbins, axnames=self.axis_names))

    @property
    def bin_lower_edges(self):
        """
        lower edges of bins. The length of the array will be nbins+1,
        since the final edge of the last bin is included for ease of
        use in vector operations
        """
        if self._bin_lower_edges is None:
            self._bin_lower_edges = [np.linspace(self._ranges[ii][0],
                                                 self._ranges[ii][-1],
                                                 self._nbins[ii] + 1)
                                     for ii in range(self.ndims)]
        return self._bin_lower_edges

    @property
    def bins(self):
        return self._nbins

    @property
    def ranges(self):
        return self._ranges

    @property
    def ndims(self):
        return len(self._nbins)

    @property
    def outliers(self):
        """
        returns the number of outlier points (the number of input
        datapoints - the sum of the histogram). This assumes the data
        of the histogram is unmodified (and represents "counts").
        """
        return self._numsamples - self.data.sum()

    def fill(self, datapoints, **kwargs):
        """
        generate a histogram from data points.  Since the results of
        fill() are added to the current histogram, you can call fill()
        multiple times to fill a single Histogram.

        Parameters
        ----------
        datapoints: array_like
            array of N-d points (see `numpy.histogramdd` documentation)
        kwargs:
            any extra options to pass to `numpy.histogramdd` when
            creating the histogram
        """

        hist, __ = np.histogramdd(datapoints, bins=self._nbins,
                                  range=self._ranges, **kwargs)

        self.data += hist
        self._numsamples += len(datapoints)

    def bin_centers(self, index):
        """
        returns array of bin centers for the given index
        """
        return 0.5 * (self.bin_lower_edges[index][1:] +
                      self.bin_lower_edges[index][0:-1])

    def to_fits(self):
        """
        Convert the `Histogram` into an `astropy.io.fits.ImageHDU`,
        suitable for writing to a file.

        Examples
        --------

        >>> myhist.to_fits().writeto("outputfile.fits.gz", clobber=True)

        """
        ohdu = fits.ImageHDU(data=self.data.transpose())
        ohdu.name = self.name
        ndim = len(self._nbins)

        # the lower-left edge of the first bin is (Xed[0],Yed[0]), which
        # is (0.5,0.5) in FITS pixel coordinates (the center of the bin is
        # at (1.0,1.0))

        # now note that this defines the first pixel in FITS coordinates
        # with the center (1.0,1.0). in integer python coordinates it is [0,0]

        # to transform a world value, you need to subtract 1.0 and round
        # down to get the bin number:
        #       ibin = round( Vpix-1.0 )
        # To get the value of ibin, you need to go the other way:
        #       Vworld[ibin] = wcs.wcs_pix2world( ibin+0.5 )

        for dim in range(ndim):
            width = self._ranges[dim][1] - self._ranges[dim][0]
            num = self._nbins[dim]
            delta = width / float(num)
            bin0pix = 0.5  # lower-left corner of first bin
            bin0coord = self._ranges[dim][0]

            name = self.axis_names[dim]

            ctype = name[0:4] + "-   "
            if self._ctypes is not None:
                ctype = self._ctypes[dim]

            ohdu.header.set("CTYPE%d" % (dim + 1), ctype, name)
            ohdu.header.set("CDELT%d" % (dim + 1), delta)
            ohdu.header.set("CRVAL%d" % (dim + 1), bin0coord)
            ohdu.header.set("CRPIX%d" % (dim + 1), bin0pix)
            ohdu.header.set("CNAME%d" % (dim + 1), self.axis_names[dim])

        if self.value_scale:
            ohdu.header.set("BSCALE", 1.0 / float(self.value_scale))

        if self.value_zero:
            ohdu.header.set("BZERO", float(self.value_zero))

        ohdu.header.set("NSAMP", self._numsamples, "Number of samples "
                                                   "originally filled")

        return ohdu

    @staticmethod
    def from_fits(input_fits):
        """
        Construct a `Histogram` from a previously written FITS histogram file
        or HDU (see `Histogram.to_fits()`)

        Parameters
        ----------
        input_fits: string or astropy.io.fits.ImageHDU
            File or HDU to read into histogram (Should be a FITS HDU
            originally created by `Histogram.to_fits()`, may not work for
            general FITS images)
        """

        hist = Histogram()

        if type(input_fits) == str:
            hdu = fits.open(input_fits)[1]
        else:
            hdu = input_fits

        hist.data = hdu.data.transpose()
        hist._nbins = hist.data.shape

        wcs = WCS(hdu.header)
        ndim = len(hist._nbins)

        edges = []
        hist._ranges = []
        axis_names = []
        hist._ctypes = []

        for dim in range(ndim):
            # note that histogramdd returns edges for 0-N+1 (including
            # the lower edge of the non-existant next bin), so we do
            # the same here to keep things the same
            ax = np.zeros((hist._nbins[dim] + 1, ndim))
            ax[:, dim] = np.arange(hist._nbins[dim] + 1) + 0.5
            edges.append(wcs.wcs_pix2world(ax, 1)[:, dim])
            hist._ranges.append((edges[dim][0], edges[dim][-1]))
            hist._ctypes.append(hdu.header["CTYPE%d" % (dim + 1)])
            if hdu.header.get("CNAME%d" % (dim + 1)):
                axis_names.append(hdu.header["CNAME%d" % (dim + 1)])
            else:
                axis_names.append(hist._ctypes[dim][0:4])

        hist.axis_names = np.array(axis_names)
        hist._bin_lower_edges = edges
        hist._ranges = np.array(hist._ranges)

        if hdu.header.get("BSCALE"):
            hist.value_scale = hdu.header["BSCALE"]

        if hdu.header.get("BSCALE"):
            hist.value_zero = hdu.header["BZERO"]

        if hdu.header.get("NSAMP"):
            hist._numsamples += int(hdu.header["NSAMP"])

        return hist

    def get_value(self, coords, outlier_value=None):
        """ Returns the values of the histogram at the given world
        coordinate(s)

        Parameters
        ----------

        coords: array_like
            array of M coordinates of dimension N (where
            the N is the histogram dimension)
        outlier_value: float or None
          value for outliers, if None, coordinates outside the edges of the
          histogram will be given the edge value

        """
        # if (self._bin_lower_edges == None):
        #     raise ValueError("Histogram is not filled")

        if np.isnan(coords).any():
            raise ValueError("Bad coordinate value")

        world = np.array(coords, ndmin=2)  # at least 2D
        ndims = len(self._nbins)

        bins = np.array([np.digitize(world[:, ii],
                                     self.bin_lower_edges[ii][1:])
                         for ii in range(ndims)])

        maxbin = np.array(self.data.shape)

        # deal with out-of-range values:
        if outlier_value is None:
            # extrapolate (simply for now, just takes edge value)
            bins[bins < 0] = 0
            for ii in range(ndims):
                bins[ii][bins[ii] >= maxbin[ii]] = maxbin[ii] - 1
        else:
            if (bins >= maxbin).any() or (bins < 0).any():
                return outlier_value

        return self.data[tuple(bins)]

    def draw_2d(self, dims=(0, 1), **kwargs):
        """draw the histogram using pcolormesh() (only works for 2D
        histograms currently)

        Parameters
        ----------
        dims: (int,int)
            indices of which dimensions to draw
        kwargs:
            arguments to pass to matplotlib `pcolormesh` command
        """
        from matplotlib import pyplot

        if self.data.ndim < 2:
            raise ValueError("Too few dimensions")

        if len(dims) != 2:
            raise ValueError("dims must be a length-2 integer array")

        pyplot.pcolormesh(self.bin_lower_edges[dims[0]],
                          self.bin_lower_edges[dims[1]],
                          self.data, **kwargs)
        pyplot.title(self.name)
        pyplot.xlabel(self.axis_names[dims[0]])
        pyplot.ylabel(self.axis_names[dims[1]])

    def draw_1d(self, dim=0, **kwargs):
        from matplotlib import pyplot

        # todo fix this to work properly with dim argument!
        pyplot.plot(self.bin_centers(dim), self.data, drawstyle='steps-mid',
                    **kwargs)

    def resample_inplace(self, nbins):
        """
        Change the shape of the histogram using an n-dimensional
        interpolation function (via `ndimage.map_coordinates`).

        Parameters
        ----------
        nbins: tuple of int
            a tuple of the new number of bins to resample_inplace
            this histogram over (e.g. if the original histogram was
            (100,100), setting bins to (200,200) would provide double
            the resolution, with the interviening bins interpolated.
        """

        oldbins = self._nbins
        # iold = np.indices(oldbins)
        inew = np.indices(nbins)

        coords = np.array([inew[X] * (oldbins[X]) / float(nbins[X])
                           for X in range(len(nbins))])

        self._nbins = nbins
        self.data = ndimage.map_coordinates(self.data, coords)
        self._bin_lower_edges = None  # need to be recalculated

    @property
    def hist(self):
        """ for backward compatibility. Use `Histogram.data` for read/write
        access"""
        return self.data



from ctapipe.calib.camera.dl0 import CameraDL0Reducer
from ctapipe.calib.camera.r1 import HessioR1Calibrator
from ctapipe.io.hessio import hessio_event_source
from ctapipe.utils import get_dataset
from numpy.testing import assert_almost_equal


def get_test_event():
    filename = get_dataset('gamma_test.simtel.gz')
    source = hessio_event_source(filename, requested_event=409,
                                 use_event_id=True)
    event = next(source)
    return event


def previous_calibration(event):
    r1 = HessioR1Calibrator(None, None)
    r1.calibrate(event)


def test_camera_dl0_reducer():
    event = get_test_event()
    previous_calibration(event)
    telid = 11

    reducer = CameraDL0Reducer(None, None)
    reducer.reduce(event)
    waveforms = event.dl0.tel[telid].pe_samples
    assert_almost_equal(waveforms[0, 0, 0], -0.091, 3)


def test_check_r1_exists():
    telid = 11
    event = get_test_event()
    previous_calibration(event)
    reducer = CameraDL0Reducer(None, None)
    assert(reducer.check_r1_exists(event, telid) is True)
    event.r1.tel[telid].pe_samples = None
    assert(reducer.check_r1_exists(event, telid) is False)



from ctapipe.image.timing_parameters import *
import numpy as np
import astropy.units as u
from numpy.testing import assert_allclose


def test_grad_fit():
    """
    Simple test that gradient fitting gives expected answers for perfect gradient
    """
    grad = 2
    intercept = 1

    timing = timing_parameters(pix_x=np.zeros(4)*u.deg, pix_y=np.arange(4)*u.deg,
                               image=np.ones(4),
                               peak_time=intercept*u.ns +grad*np.arange(4)*u.ns,
                               rotation_angle=0*u.deg)

    # Test we get the values we put in back out again
    assert_allclose(timing.gradient,grad*u.ns/u.deg)
    assert_allclose(timing.intercept,intercept*u.deg)

    # Then try a different rotation angle
    rot_angle = 20*u.deg
    timing_rot20 = timing_parameters(pix_x=np.zeros(4)*u.deg, pix_y=np.arange(4)*u.deg,
                                     image=np.ones(4),
                                     peak_time=intercept*u.ns +grad*np.arange(4)*u.ns,
                                     rotation_angle=rot_angle)
    # Test the output again makes sense
    assert_allclose(timing_rot20.gradient, timing.gradient/np.cos(rot_angle))
    assert_allclose(timing_rot20.intercept, timing.intercept)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Utilities for reading or working with Camera geometry files
"""
import logging
from collections import defaultdict

import numpy as np
from astropy import units as u
from astropy.coordinates import Angle
from astropy.table import Table
from astropy.utils import lazyproperty
from scipy.spatial import cKDTree as KDTree

from ctapipe.utils import get_dataset, find_all_matching_datasets
from ctapipe.utils.linalg import rotation_matrix_2d
from ctapipe.core import Provenance


__all__ = ['CameraGeometry',]

logger = logging.getLogger(__name__)


# dictionary to convert number of pixels to camera + the focal length of the
# telescope into a camera type for use in `CameraGeometry.guess()`
#     Key = (num_pix, focal_length_in_meters)
#     Value = (type, subtype, pixtype, pixrotation, camrotation)
_CAMERA_GEOMETRY_TABLE = {
    (2048, 2.3): ('SST', 'CHEC', 'rectangular', 0 * u.degree, 0 * u.degree),
    (2048, 2.2): ('SST', 'CHEC', 'rectangular', 0 * u.degree, 0 * u.degree),
    (2048, 36.0): ('LST', 'HESS-II', 'hexagonal', 0 * u.degree,
                   0 * u.degree),
    (960, None): ('MST', 'HESS-I', 'hexagonal', 0 * u.degree,
                  0 * u.degree),
    (1855, 16.0): ('MST', 'NectarCam', 'hexagonal',
                   0 * u.degree, -100.893 * u.degree),
    (1855, 28.0): ('LST', 'LSTCam', 'hexagonal',
                   0. * u.degree, -100.893 * u.degree),
    (1296, None): ('SST', 'DigiCam', 'hexagonal', 30 * u.degree, 0 * u.degree),
    (1764, None): ('MST', 'FlashCam', 'hexagonal', 30 * u.degree, 0 * u.degree),
    (2368, None): ('SST', 'ASTRICam', 'rectangular', 0 * u.degree,
                   0 * u.degree),
    (11328, None): ('SCT', 'SCTCam', 'rectangular', 0 * u.degree, 0 * u.degree),
}


class CameraGeometry:
    """`CameraGeometry` is a class that stores information about a
    Cherenkov Camera that us useful for imaging algorithms and
    displays. It contains lists of pixel positions, areas, pixel
    shapes, as well as a neighbor (adjacency) list and matrix for each pixel. 
    In general the neighbor_matrix attribute should be used in any algorithm 
    needing pixel neighbors, since it is much faster. See for example 
    `ctapipe.image.tailcuts_clean` 

    The class is intended to be generic, and work with any Cherenkov
    Camera geometry, including those that have square vs hexagonal
    pixels, gaps between pixels, etc.

    You can construct a CameraGeometry either by specifying all data, 
    or using the `CameraGeometry.guess()` constructor, which takes metadata 
    like the pixel positions and telescope focal length to look up the rest 
    of the data. Note that this function is memoized, so calling it multiple 
    times with the same inputs will give back the same object (for speed).

    Parameters
    ----------
    self: type
        description
    cam_id: camera id name or number
        camera identification string
    pix_id: array(int)
        pixels id numbers
    pix_x: array with units
        position of each pixel (x-coordinate)
    pix_y: array with units
        position of each pixel (y-coordinate)
    pix_area: array(float)
        surface area of each pixel, if None will be calculated
    neighbors: list(arrays)
        adjacency list for each pixel
    pix_type: string
        either 'rectangular' or 'hexagonal'
    pix_rotation: value convertable to an `astropy.coordinates.Angle`
        rotation angle with unit (e.g. 12 * u.deg), or "12d"
    cam_rotation: overall camera rotation with units
    """

    _geometry_cache = {}  # dictionary CameraGeometry instances for speed

    def __init__(self, cam_id, pix_id, pix_x, pix_y, pix_area, pix_type,
                 pix_rotation="0d", cam_rotation="0d",
                 neighbors=None, apply_derotation=True):

        self.cam_id = cam_id
        self.pix_id = pix_id
        self.pix_x = pix_x
        self.pix_y = pix_y
        self.pix_area = pix_area
        self.pix_type = pix_type
        self.pix_rotation = Angle(pix_rotation)
        self.cam_rotation = Angle(cam_rotation)
        self._precalculated_neighbors = neighbors

        if self.pix_area is None:
            self.pix_area = CameraGeometry._calc_pixel_area(pix_x, pix_y,
                                                            pix_type)

        if apply_derotation:
            # todo: this should probably not be done, but need to fix
            # GeometryConverter and reco algorithms if we change it.
            if len(pix_x.shape) == 1:
                self.rotate(cam_rotation)

    def __eq__(self, other):
        return ((self.cam_id == other.cam_id)
                and (self.pix_x == other.pix_x).all()
                and (self.pix_y == other.pix_y).all()
                and (self.pix_type == other.pix_type)
                and (self.pix_rotation == other.pix_rotation)
                and (self.pix_type == other.pix_type)
                )

    @classmethod
    @u.quantity_input
    def guess(cls, pix_x: u.m, pix_y: u.m, optical_foclen: u.m,
              apply_derotation=True):
        """ 
        Construct a `CameraGeometry` by guessing the appropriate quantities
        from a list of pixel positions and the focal length. 
        """
        # only construct a new one if it has never been constructed before,
        # to speed up access. Otherwise return the already constructed instance
        # the identifier uses the values of pix_x (which are converted to a
        # string to make them hashable) and the optical_foclen. So far,
        # that is enough to uniquely identify a geometry.
        identifier = (pix_x.value.tostring(), optical_foclen)
        if identifier in CameraGeometry._geometry_cache:
            return CameraGeometry._geometry_cache[identifier]

        # now try to determine the camera type using the map defined at the
        # top of this file.

        tel_type, cam_id, pix_type, pix_rotation, cam_rotation = \
            _guess_camera_type(len(pix_x), optical_foclen)

        area = cls._calc_pixel_area(pix_x, pix_y, pix_type)

        instance = cls(
            cam_id=cam_id,
            pix_id=np.arange(len(pix_x)),
            pix_x=pix_x,
            pix_y=pix_y,
            pix_area=np.ones(pix_x.shape) * area,
            neighbors=None,
            pix_type=pix_type,
            pix_rotation=Angle(pix_rotation),
            cam_rotation=Angle(cam_rotation),
            apply_derotation=apply_derotation
        )

        CameraGeometry._geometry_cache[identifier] = instance
        return instance

    @staticmethod
    def _calc_pixel_area(pix_x, pix_y, pix_type):
        """ recalculate pixel area based on the pixel type and layout

        Note this will not work on cameras with varying pixel sizes.
        """

        dist = _get_min_pixel_seperation(pix_x, pix_y)

        if pix_type.startswith('hex'):
            rad = dist / np.sqrt(3)  # radius to vertex of hexagon
            area = rad ** 2 * (3 * np.sqrt(3) / 2.0)  # area of hexagon
        elif pix_type.startswith('rect'):
            area = dist ** 2
        else:
            raise KeyError("unsupported pixel type")

        return np.ones(pix_x.shape) * area

    @classmethod
    def get_known_camera_names(cls, array_id='CTA'):
        """
        Returns a list of camera_ids that are registered in 
        `ctapipe_resources`. These are all the camera-ids that can be 
        instantiated by the `from_name` method

        Parameters
        ----------
        array_id: str 
            which array to search (default CTA)

        Returns
        -------
        list(str)
        """

        pattern = "(.*)\.camgeom\.fits(\.gz)?"
        return find_all_matching_datasets(pattern, regexp_group=1)

    @classmethod
    def from_name(cls, camera_id='NectarCam', version=None):
        """
        Construct a CameraGeometry using the name of the camera and array.

        This expects that there is a resource in the `ctapipe_resources` module
        called "[array]-[camera].camgeom.fits.gz" or "[array]-[camera]-[
        version].camgeom.fits.gz"

        Parameters
        ----------
        camera_id: str
           name of camera (e.g. 'NectarCam', 'LSTCam', 'GCT', 'SST-1M')
        array_id: str
           array identifier (e.g. 'CTA', 'HESS')
        version:
           camera version id (currently unused)

        Returns
        -------
        new CameraGeometry
        """

        if version is None:
            verstr = ''
        else:
            verstr = "-{:03d}".format(version)

        filename = get_dataset("{camera_id}{verstr}.camgeom.fits.gz"
                               .format(camera_id=camera_id, verstr=verstr))
        Provenance().add_input_file(filename, role='dl0.tel.svc.camera')
        return CameraGeometry.from_table(filename)

    def to_table(self):
        """ convert this to an `astropy.table.Table` """
        # currently the neighbor list is not supported, since
        # var-length arrays are not supported by astropy.table.Table
        return Table([self.pix_id, self.pix_x, self.pix_y, self.pix_area],
                     names=['pix_id', 'pix_x', 'pix_y', 'pix_area'],
                     meta=dict(PIX_TYPE=self.pix_type,
                               TAB_TYPE='ctapipe.instrument.CameraGeometry',
                               TAB_VER='1.0',
                               CAM_ID=self.cam_id,
                               PIX_ROT=self.pix_rotation.deg,
                               CAM_ROT=self.cam_rotation.deg,
                               ))

    @classmethod
    def from_table(cls, url_or_table, **kwargs):
        """
        Load a CameraGeometry from an `astropy.table.Table` instance or a 
        file that is readable by `astropy.table.Table.read()`

        Parameters
        ----------
        url_or_table: string or astropy.table.Table
            either input filename/url or a Table instance

        format: str
            astropy.table format string (e.g. 'ascii.ecsv') in case the 
            format cannot be determined from the file extension

        kwargs: extra keyword arguments
            extra arguments passed to `astropy.table.read()`, depending on 
            file type (e.g. format, hdu, path)


        """

        tab = url_or_table
        if not isinstance(url_or_table, Table):
            tab = Table.read(url_or_table, **kwargs)

        return cls(
            cam_id=tab.meta.get('CAM_ID', 'Unknown'),
            pix_id=tab['pix_id'],
            pix_x=tab['pix_x'].quantity,
            pix_y=tab['pix_y'].quantity,
            pix_area=tab['pix_area'].quantity,
            pix_type=tab.meta['PIX_TYPE'],
            pix_rotation=Angle(tab.meta['PIX_ROT'] * u.deg),
            cam_rotation=Angle(tab.meta['CAM_ROT'] * u.deg),
        )

    def __repr__(self):
        return "CameraGeometry(cam_id='{cam_id}', pix_type='{pix_type}', " \
               "npix={npix}, cam_rot={camrot}, pix_rot={pixrot})".format(
                   cam_id=self.cam_id,
                   pix_type=self.pix_type,
                   npix=len(self.pix_id),
                   pixrot=self.pix_rotation,
                   camrot=self.cam_rotation
               )

    def __str__(self):
        return self.cam_id

    @lazyproperty
    def neighbors(self):
        """" only calculate neighbors when needed or if not already 
        calculated"""

        # return pre-calculated ones (e.g. those that were passed in during
        # the object construction) if they exist
        if self._precalculated_neighbors is not None:
            return self._precalculated_neighbors

        # otherwise compute the neighbors from the pixel list
        dist = _get_min_pixel_seperation(self.pix_x, self.pix_y)

        neighbors = _find_neighbor_pixels(
            self.pix_x.value,
            self.pix_y.value,
            rad=1.4 * dist.value
        )

        return neighbors

    @lazyproperty
    def neighbor_matrix(self):
        return _neighbor_list_to_matrix(self.neighbors)

    def rotate(self, angle):
        """rotate the camera coordinates about the center of the camera by
        specified angle. Modifies the CameraGeometry in-place (so
        after this is called, the pix_x and pix_y arrays are
        rotated.

        Notes
        -----

        This is intended only to correct simulated data that are
        rotated by a fixed angle.  For the more general case of
        correction for camera pointing errors (rotations,
        translations, skews, etc), you should use a true coordinate
        transformation defined in `ctapipe.coordinates`.

        Parameters
        ----------

        angle: value convertable to an `astropy.coordinates.Angle`
            rotation angle with unit (e.g. 12 * u.deg), or "12d"

        """
        rotmat = rotation_matrix_2d(angle)
        rotated = np.dot(rotmat.T, [self.pix_x.value, self.pix_y.value])
        self.pix_x = rotated[0] * self.pix_x.unit
        self.pix_y = rotated[1] * self.pix_x.unit
        self.pix_rotation -= Angle(angle)
        self.cam_rotation -= Angle(angle)

    @classmethod
    def make_rectangular(cls, npix_x=40, npix_y=40, range_x=(-0.5, 0.5),
                         range_y=(-0.5, 0.5)):
        """Generate a simple camera with 2D rectangular geometry.

        Used for testing.

        Parameters
        ----------
        npix_x : int
            number of pixels in X-dimension
        npix_y : int
            number of pixels in Y-dimension
        range_x : (float,float)
            min and max of x pixel coordinates in meters
        range_y : (float,float)
            min and max of y pixel coordinates in meters

        Returns
        -------
        CameraGeometry object

        """
        bx = np.linspace(range_x[0], range_x[1], npix_x)
        by = np.linspace(range_y[0], range_y[1], npix_y)
        xx, yy = np.meshgrid(bx, by)
        xx = xx.ravel() * u.m
        yy = yy.ravel() * u.m

        ids = np.arange(npix_x * npix_y)
        rr = np.ones_like(xx).value * (xx[1] - xx[0]) / 2.0

        return cls(cam_id=-1,
                   pix_id=ids,
                   pix_x=xx * u.m,
                   pix_y=yy * u.m,
                   pix_area=(2 * rr) ** 2,
                   neighbors=None,
                   pix_type='rectangular')


# ======================================================================
# utility functions:
# ======================================================================

def _get_min_pixel_seperation(pix_x, pix_y):
    """
    Obtain the minimum seperation between two pixels on the camera

    Parameters
    ----------
    pix_x : array_like
        x position of each pixel
    pix_y : array_like
        y position of each pixels

    Returns
    -------
    pixsep : astropy.units.Unit

    """
    #    dx = pix_x[1] - pix_x[0]    <=== Not adjacent for DC-SSTs!!
    #    dy = pix_y[1] - pix_y[0]

    dx = pix_x - pix_x[0]
    dy = pix_y - pix_y[0]
    pixsep = np.min(np.sqrt(dx ** 2 + dy ** 2)[1:])
    return pixsep


def _find_neighbor_pixels(pix_x, pix_y, rad):
    """use a KD-Tree to quickly find nearest neighbors of the pixels in a
    camera. This function can be used to find the neighbor pixels if
    they are not already present in a camera geometry file.

    Parameters
    ----------
    pix_x : array_like
        x position of each pixel
    pix_y : array_like
        y position of each pixels
    rad : float
        radius to consider neighbor it should be slightly larger
        than the pixel diameter.

    Returns
    -------
    array of neighbor indices in a list for each pixel

    """

    points = np.array([pix_x, pix_y]).T
    indices = np.arange(len(pix_x))
    kdtree = KDTree(points)
    neighbors = [kdtree.query_ball_point(p, r=rad) for p in points]
    for nn, ii in zip(neighbors, indices):
        nn.remove(ii)  # get rid of the pixel itself
    return neighbors


def _guess_camera_type(npix, optical_foclen):
    global _CAMERA_GEOMETRY_TABLE

    try:
        return _CAMERA_GEOMETRY_TABLE[(npix, None)]
    except KeyError:
        return _CAMERA_GEOMETRY_TABLE.get((npix, round(optical_foclen.value, 1)),
                                          ('unknown', 'unknown', 'hexagonal',
                                           0 * u.degree, 0 * u.degree))


def _neighbor_list_to_matrix(neighbors):
    """ 
    convert a neighbor adjacency list (list of list of neighbors) to a 2D 
    numpy array, which is much faster (and can simply be multiplied)
    """

    npix = len(neighbors)
    neigh2d = np.zeros(shape=(npix, npix), dtype=np.bool)

    for ipix, neighbors in enumerate(neighbors):
        for jn, neighbor in enumerate(neighbors):
            neigh2d[ipix, neighbor] = True

    return neigh2d






import pytest
from traitlets import Float, TraitError

from .. import Tool


def test_tool_simple():
    """test the very basic functionality of a Tool"""

    class MyTool(Tool):
        description = "test"
        userparam = Float(5.0, help="parameter").tag(config=True)

    tool = MyTool()
    tool.userparam = 1.0
    tool.log_level = 'DEBUG'
    tool.log.info("test")
    tool.run([])

    # test parameters changes:
    tool.userparam = 4.0
    with pytest.raises(TraitError):
        tool.userparam = "badvalue"


def test_tool_version():

    class MyTool(Tool):
        description = "test"
        userparam = Float(5.0, help="parameter").tag(config=True)

    tool = MyTool()
    assert tool.version_string != ""



from ctapipe.io.hessio import hessio_event_source
from ctapipe.utils import get_dataset


def test_get_run_id():
    filename = get_dataset("gamma_test.simtel.gz")
    print(filename)
    gen = hessio_event_source(filename)
    event = next(gen)
    tels = event.dl0.tels_with_data
    print(tels)
    assert tels == {38, 47}


def test_get_specific_event():
    dataset = get_dataset("gamma_test.simtel.gz")
    source = hessio_event_source(dataset, requested_event=2)
    event = next(source)
    assert event.count == 2
    assert event.dl0.event_id == 803
    source = hessio_event_source(dataset, requested_event=803,
                                 use_event_id=True)
    event = next(source)
    assert event.count == 2
    assert event.dl0.event_id == 803



"""
Calibrator for the DL0 -> DL1 data level transition.

This module handles the calibration from the DL0 data level to DL1. This
transition involves the waveform cleaning (such as filtering, smoothing,
or basline subtraction) performed by a cleaner from
`ctapipe.image.waveform_cleaning`, and the charge extraction technique
from `ctapipe.image.charge_extractors`.
"""
import numpy as np

from ctapipe.core import Component
from ctapipe.core.traits import Float
from ctapipe.image import NeighbourPeakIntegrator, NullWaveformCleaner
from ctapipe.instrument import CameraGeometry

__all__ = ['CameraDL1Calibrator']


def integration_correction(n_chan, pulse_shape, refstep, time_slice,
                           window_width, window_shift):
    """
    Obtain the integration correction for the window specified.

    This correction accounts for the cherenkov signal that may be missed due
    to a smaller integration window by looking at the reference pulse shape.

    Provides the same result as set_integration_correction from readhess.

    Parameters
    ----------
    n_chan : int
        Number of gain channels for the telescope
    pulse_shape : ndarray
        Numpy array containing the pulse shape for each channel.
    refstep : int
        The step in time for each sample of the reference pulse shape
    time_slice : int
        The step in time for each sample of the waveforms
    window_width : int
        Width of the integration window.
    window_shift : int
        Shift to before the peak for the start of the integration window.

    Returns
    -------
    correction : list[2]
        Value of the integration correction for this telescope for each
        channel.
    """
    correction = np.ones(n_chan)
    for chan in range(n_chan):
        pshape = pulse_shape[chan]
        if pshape.all() is False or time_slice == 0 or refstep == 0:
            continue

        ref_x = np.arange(0, pshape.size * refstep, refstep)
        edges = np.arange(0, pshape.size * refstep + 1, time_slice)

        sampled, se = np.histogram(ref_x, edges, weights=pshape, density=True)
        n_samples = sampled.size
        start = sampled.argmax() - window_shift
        end = start + window_width

        if window_width > n_samples:
            window_width = n_samples
        if start < 0:
            start = 0
        if start + window_width > n_samples:
            start = n_samples - window_width

        integration = np.diff(se)[start:end] * sampled[start:end]
        correction[chan] = 1 / np.sum(integration)

    return correction


class CameraDL1Calibrator(Component):
    """
    The calibrator for DL1 charge extraction. Fills the dl1 container.

    It handles the integration correction and, if required, the list of
    neighbours.

    Parameters
    ----------
    config : traitlets.loader.Config
        Configuration specified by config file or cmdline arguments.
        Used to set traitlet values.
        Set to None if no configuration to pass.
    tool : ctapipe.core.Tool or None
        Tool executable that is calling this component.
        Passes the correct logger to the component.
        Set to None if no Tool to pass.
    extractor : ctapipe.calib.camera.charge_extractors.ChargeExtractor
        The extractor to use to extract the charge from the waveforms.
        By default the NeighbourPeakIntegrator with default configuration
        is used.
    cleaner : ctapipe.calib.camera.waveform_cleaners.Cleaner
        The waveform cleaner to use. By default no cleaning is
        applied to the waveforms.
    kwargs
    """

    name = 'CameraCalibrator'
    radius = Float(None, allow_none=True,
                   help='Pixels within radius from a pixel are considered '
                        'neighbours to the pixel. Set to None for the default '
                        '(1.4 * min_pixel_seperation).').tag(config=True)
    clip_amplitude = Float(None, allow_none=True,
                           help='Amplitude in p.e. above which the signal is '
                                'clipped. Set to None for no '
                                'clipping.').tag(config=True)

    def __init__(self, config, tool, extractor=None, cleaner=None, **kwargs):
        super().__init__(config=config, parent=tool, **kwargs)
        self.extractor = extractor
        if self.extractor is None:
            self.extractor = NeighbourPeakIntegrator(config, tool)
        self.cleaner = cleaner
        if self.cleaner is None:
            self.cleaner = NullWaveformCleaner(config, tool)
        self._dl0_empty_warn = False

    def check_dl0_exists(self, event, telid):
        """
        Check that dl0 data exists. If it does not, then do not change dl1.

        This ensures that if the containers were filled from a file containing
        dl1 data, it is not overwritten by non-existant data.

        Parameters
        ----------
        event : container
            A `ctapipe` event container
        telid : int
            The telescope id.

        Returns
        -------
        bool
            True if dl0.tel[telid].pe_samples is not None, else false.
        """
        dl0 = event.dl0.tel[telid].pe_samples
        if dl0 is not None:
            return True
        else:
            if not self._dl0_empty_warn:
                self.log.warning("Encountered an event with no DL0 data. "
                                 "DL1 is unchanged in this circumstance.")
                self._dl0_empty_warn = True
            return False

    @staticmethod
    def get_geometry(event, telid):
        """
        Obtain the geometry for this telescope.

        Parameters
        ----------
        event : container
            A `ctapipe` event container
        telid : int
            The telescope id.
            The neighbours are calculated once per telescope.

        Returns
        -------
        `CameraGeometry`
        """
        return CameraGeometry.guess(*event.inst.pixel_pos[telid],
                                    event.inst.optical_foclen[telid])

    def get_correction(self, event, telid):
        """
        Obtain the integration correction for this telescope.

        Parameters
        ----------
        event : container
            A `ctapipe` event container
        telid : int
            The telescope id.
            The integration correction is calculated once per telescope.

        Returns
        -------
        ndarray
        """
        try:
            shift = self.extractor.window_shift
            width = self.extractor.window_width
            n_chan = event.inst.num_channels[telid]
            shape = event.mc.tel[telid].reference_pulse_shape
            step = event.mc.tel[telid].meta['refstep']
            time_slice = event.mc.tel[telid].time_slice
            correction = integration_correction(n_chan, shape, step,
                                                time_slice, width, shift)
            return correction
        except AttributeError:
            # Don't apply correction when window_shift or window_width
            # does not exist in extractor, or when container does not have
            # a reference pulse shape
            return np.ones(event.inst.num_channels[telid])

    def calibrate(self, event):
        """
        Fill the dl1 container with the calibration data that results from the
        configuration of this calibrator.

        Parameters
        ----------
        event : container
            A `ctapipe` event container
        """
        for telid in event.dl0.tels_with_data:

            if self.check_dl0_exists(event, telid):
                waveforms = event.dl0.tel[telid].pe_samples

                # Clean waveforms
                cleaned = self.cleaner.apply(waveforms)

                # Extract charge
                if self.extractor.requires_neighbours():
                    e = self.extractor
                    e.neighbours = self.get_geometry(event, telid).neighbors
                extract = self.extractor.extract_charge
                charge, peakpos, window = extract(cleaned)

                # Apply integration correction
                corrected = charge * self.get_correction(event, telid)[:, None]

                # Clip amplitude
                if self.clip_amplitude:
                    corrected[corrected > self.clip_amplitude] = \
                        self.clip_amplitude

                # Store into event container
                event.dl1.tel[telid].image = corrected
                event.dl1.tel[telid].extracted_samples = window
                event.dl1.tel[telid].peakpos = peakpos
                event.dl1.tel[telid].cleaned = cleaned



"""
Class for performing a HESS style 2D fit of muon images

To do:
    - Deal with astropy untis better, currently stripped and no checks made
    - unit tests
    - create container class for output

"""
import numpy as np
from scipy.ndimage.filters import correlate1d
from iminuit import Minuit
from scipy.optimize import minimize
from astropy import units as u
from astropy.constants import alpha
from ctapipe.io.containers import MuonIntensityParameter
from scipy.stats import norm

__all__ = ['MuonLineIntegrate']


class MuonLineIntegrate:
    """
    Object for calculating the expected 2D shape of muon image for a given
    mirror geometry. Geometry is passed to the class as
    a series of points defining the outer edge of the array
    and a set of points defining the shape of the central hole (if present).
    Muon profiles are then calculated bsed of the line integral along a given
    axis from the muon impact point.

    Expected 2D images can then be generated when the pixel geometry
    is passed to the class.

    Parameters
    ----------
    mirror_radius: float
        Radius of telescope mirror (circular approx)
    hole_radius: float
        Radius of telescope mirror hole (circular approx)
    pixel_width: float
        width of pixel in camera
    oversample_bins: int
        number of angular bins to evaluate for each pixel width
    sct_flag: bool
        flags whether the telescope uses Schwarzschild-Couder Optics
    secondary_radius: float
        Radius of the secondary mirror (circular approx) for dual
        mirror telescopes
    minlambda: float
        minimum wavelength for integration (300nm typical)
    maxlambda: float
        maximum wavelength for integration (600nm typical)
    photemit: float
        1/lambda^2 integrated over the above defined wavelength range
        multiplied by the fine structure constant (1/137)
    """
    def __init__(self, mirror_radius, hole_radius, pixel_width=0.2,
                 oversample_bins=3,sct_flag=False,secondary_radius=1.):

        self.mirror_radius = mirror_radius
        self.hole_radius = hole_radius
        self.pixel_width = pixel_width
        self.oversample_bins = oversample_bins
        self.sct_flag = sct_flag
        self.secondary_radius = secondary_radius
        self.pixel_x = 0
        self.pixel_y = 0
        self.image = 0
        self.prediction = 0
        self.minlambda = 300.e-9
        self.maxlambda = 600.e-9
        self.photemit = alpha*(self.minlambda**-1 - self.maxlambda**-1)#12165.45
        self.unit = u.deg

    @staticmethod
    def chord_length(radius, rho, phi):
        """
        Function for integrating the length of a chord across a circle

        Parameters
        ----------
        radius: float
            radius of circle
        rho: float
            fractional distance of impact point from array centre
        phi: ndarray
            rotation angles to calculate length

        Returns
        -------
        ndarray: chord length
        """
        chord = 1 - (rho * rho * np.sin(phi) * np.sin(phi))

        if rho <= 1:
            chord = radius * (np.sqrt(chord) + rho * np.cos(phi))
        elif rho > 1:
            chord = 2. * radius * np.sqrt(chord)

        chord[np.isnan(chord)] = 0
        chord[chord < 0] = 0

        return chord

    def intersect_circle(self, r, angle):
        """Perform line integration along a given axis in the mirror frame
        given an impact point on the mirrot

        Parameters
        ----------
        impact_x: float
            Impact position on mirror (tilted telescope system)
        impact_y: float
            Impact position on mirror (tilted telescope system)
        angle: float
            Angle along which to integrate mirror

        Returns
        --------
        float: length from impact point to mirror edge

        """
        mirror_length = self.chord_length(
            self.mirror_radius, r / self.mirror_radius.value, angle
        )
        hole_length = 0 * mirror_length#.unit
        if self.hole_radius > 0:
            hole_length = self.chord_length(
                self.hole_radius, r / self.hole_radius.value, angle
            )

        if self.sct_flag:
            self.sct_flag = False #Do not treat differently for now...work in progress

        if self.sct_flag:
            secondary_length = self.chord_length(
                self.secondary_radius, r / self.secondary_radius, angle
            )
            #Should be areas not lengths here?
            factor = mirror_length - (secondary_length)
            factor /= (mirror_length - hole_length)

        if not self.sct_flag:
            return mirror_length-hole_length
        else:
            return (mirror_length - hole_length + secondary_length)

    def plot_pos(self, impact_parameter, radius, phi):
        """
        Perform intersection over all angles and return length

        Parameters
        ----------
        impact_parameter: float
            Impact distance from mirror centre
        ang: ndarray
            Angles over which to integrate
        phi: float
            Rotation angle of muon image
        
        Returns
        -------
        ndarray
            Chord length for each angle
        """

        bins = int((2 * np.pi * radius) / self.pixel_width.value) * self.oversample_bins
        #ang = np.linspace(-np.pi * u.rad + phi, np.pi * u.rad + phi, bins)
        ang = np.linspace(-np.pi + phi, np.pi + phi, bins)
        l = self.intersect_circle(impact_parameter, ang)
        l = correlate1d(l, np.ones(self.oversample_bins), mode='wrap', axis=0)
        l /= self.oversample_bins

        return ang, l

    def pos_to_angle(self, centre_x, centre_y, pixel_x, pixel_y):
        """
        Convert pixel positions from x,y coordinates to rotation angle

        Parameters
        ----------
        centre_x: float
            Reconstructed image centre
        centre_y: float
            Reconstructed image centre
        pixel_x: ndarray
            Pixel x position
        pixel_y: ndarray
            Pixel y position

        Returns
        -------
         ndarray:
            Pixel rotation angle

        """
        del_x = pixel_x - centre_x
        del_y = pixel_y - centre_y

        ang = np.arctan2(del_x, del_y)
        return ang

    def image_prediction( self, impact_parameter, phi, centre_x,
                          centre_y, radius, ring_width, pixel_x, pixel_y, ):
        """Function for producing the expected image for a given set of trial
        muon parameters

        Parameters
        ----------
        impact_parameter: float
            Impact distance of muon
        centre_x: float
            Muon ring centre in field of view
        centre_y: float
            Muon ring centre in field of view
        radius: float
            Radius of muon ring
        ring_width: float
            Gaussian width of muon ring
        pixel_x: ndarray
            Pixel x coordinate
        pixel_y: ndarray
            Pixel y coordinate

        Returns
        -------
        ndarray:
            Predicted signal

        """

        # First produce angular position of each pixel w.r.t muon centre
        ang = self.pos_to_angle(centre_x, centre_y, pixel_x, pixel_y)
        # Add muon rotation angle
        ang += phi
        # Produce smoothed muon profile

        ang_prof, profile = self.plot_pos(impact_parameter, radius, phi)
        # Produce gaussian weight for each pixel give ring width
        radial_dist = np.sqrt((pixel_x - centre_x)**2 + (pixel_y - centre_y)**2)
        gauss = norm.pdf(radial_dist, radius, ring_width)

        # interpolate profile to find prediction for each pixel
        pred = np.interp(ang, ang_prof, profile)

        # Multiply by integrated emissivity between 300 and 600 nm
        pred *= 0.5 * self.photemit

        # weight by pixel width
        pred *= (self.pixel_width / radius)
        pred *= np.sin(2 * radius)
        # weight by gaussian width
        pred *= self.pixel_width*gauss

        return pred

    def likelihood(self, impact_parameter, phi, centre_x, centre_y,
                   radius, ring_width, optical_efficiency_muon):
        """
        Likelihood function to be called by minimiser

        Parameters
        ----------
        impact_parameter: float
            Impact distance from telescope centre
        centre_x: float
            Centre of muon ring in FoV
        centre_y: float
            Centre of muon ring in FoV
        radius: float
            Radius of muon ring
        ring_width: float
            Gaussian width of muon ring
        optical_efficiency_muon: float
            Efficiency of the optical system

        Returns
        -------
        float: Likelihood that model matches data
        """
        #centre_x *= self.unit
        #centre_y *= self.unit
        #radius *= self.unit
        #ring_width *= self.unit
        #impact_parameter *= u.m
        #phi *= u.rad

        # Generate model prediction
        self.prediction = self.image_prediction(
            impact_parameter,
            phi,
            centre_x,
            centre_y,
            radius,
            ring_width,
            self.pixel_x.value,
            self.pixel_y.value,
        )
        #TEST: extra scaling factor, HESS style (ang pix size squared/2piR)
        #Not including bins atm
        #HESSscale = self.pixel_width*self.pixel_width / (2.*np.pi() * radius)
        #prediction *= HESSscale

        # scale prediction by optical efficiency of array
        self.prediction *= optical_efficiency_muon

        # Multiply sum of likelihoods by -2 to make them behave like chi-squared
        return -2 * np.sum(self.calc_likelihood(self.image, self.prediction, 0.5, 1.1))

    @staticmethod
    def calc_likelihood(image, pred, spe_width, ped):
        """Calculate likelihood of prediction given the measured signal,
        gaussian approx from [denaurois2009]_

        Parameters
        ----------
        image: ndarray
            Pixel amplitudes from image
        pred: ndarray
            Predicted pixel amplitudes from model
        spe_width: ndarray
            width of single p.e. distributio
        ped: ndarray
            width of pedestal

        Returns
        -------
        ndarray: likelihood for each pixel

        """
        sq = 1 / np.sqrt(2 * np.pi * (ped**2 + pred * (1 + spe_width**2)))
        diff = (image - pred)**2
        denom = 2 * (ped**2 + pred * (1 + spe_width**2))
        expo = np.exp(-diff / denom)
        sm = expo < 1e-300
        expo[sm] = 1e-300
        return np.log(sq * expo)

    def fit_muon(self, centre_x, centre_y, radius, pixel_x, pixel_y, image):
        """

        Parameters
        ----------
        centre_x: float
            Centre of muon ring in the field of view from circle fitting
        centre_y: float
            Centre of muon ring in the field of view from circle fitting
        radius: float
            Radius of muon ring from circle fitting
        pixel_x: ndarray
            X position of pixels in image from circle fitting
        pixel_y: ndarray
            Y position of pixel in image from circle fitting
        image: ndarray
            Amplitude of image pixels

        Returns
        -------
        MuonIntensityParameters
        """

        # First store these parameters in the class so we can use them in minimisation
        self.image = image
        self.pixel_x = pixel_x.to(u.deg)
        self.pixel_y = pixel_y.to(u.deg)
        self.unit = pixel_x.unit

        radius.to(u.deg)
        centre_x.to(u.deg)
        centre_y.to(u.deg)

        # Return interesting stuff
        fitoutput = MuonIntensityParameter()

        init_params = {}
        init_errs = {}
        init_constrain = {}
        init_params['impact_parameter'] = 4.
        init_params['phi'] = 0.
        init_params['radius'] = radius.value
        init_params['centre_x'] = centre_x.value
        init_params['centre_y'] = centre_y.value
        init_params['ring_width'] = 0.1
        init_params['optical_efficiency_muon'] = 0.1
        init_errs['error_impact_parameter'] = 2.
        init_constrain['limit_impact_parameter'] = (0.,25.)
        init_errs['error_phi'] = 0.1
        init_errs['error_ring_width'] = 0.001 * radius.value
        init_errs['error_optical_efficiency_muon'] = 0.05
        init_constrain['limit_phi'] = (-np.pi,np.pi)
        init_constrain['fix_radius'] = True
        init_constrain['fix_centre_x'] = True
        init_constrain['fix_centre_y'] = True
        init_constrain['limit_ring_width'] = (0.,1.)
        init_constrain['limit_optical_efficiency_muon'] = (0.,1.)

        print("radius =",radius," pre migrad")

        parameter_names = init_params.keys()

        # Create Minuit object with first guesses at parameters
        # strip away the units as Minuit doesnt like them
        minuit = Minuit(
            self.likelihood,
            #forced_parameters=parameter_names,
            **init_params,
            **init_errs,
            **init_constrain,
            errordef=1.,
            print_level=1
            #pedantic=False
        )

        # Perform minimisation
        minuit.migrad()
            
        # Get fitted values
        fit_params = minuit.values
        fitoutput.impact_parameter = fit_params['impact_parameter']*u.m
        #fitoutput.phi = fit_params['phi']*u.rad
        fitoutput.impact_parameter_pos_x = fit_params['impact_parameter'] * np.cos(fit_params['phi'] * u.rad) * u.m
        fitoutput.impact_parameter_pos_y = fit_params['impact_parameter'] * np.sin(fit_params['phi'] * u.rad) * u.m
        fitoutput.ring_width = fit_params['ring_width']*self.unit
        fitoutput.optical_efficiency_muon = fit_params['optical_efficiency_muon']

        fitoutput.prediction = self.prediction

        return fitoutput



"""
Dump instrumental descriptions in a monte-carlo (simtelarray) input file to 
FITS files that can be loaded independently (e.g. with 
CameraGeometry.from_table()).  The name of the output files are 
automatically generated.
"""

from collections import defaultdict

import numpy as np
from astropy import units as u
from astropy.table import Table

from ctapipe.core import Tool, Provenance
from ctapipe.core.traits import (Unicode, Dict, Enum)
from ctapipe.io.hessio import hessio_event_source


def get_camera_types(subarray):
    """ return dict of camera names mapped to a list of tel_ids
     that use that camera

     Parameters
     ----------
     subarray: ctapipe.instrument.SubarrayDescription

     """

    cam_types = defaultdict(list)

    for telid in subarray.tel:
        geom = subarray.tel[telid].camera
        cam_types[geom.cam_id].append(telid)

    return cam_types


class DumpInstrumentTool(Tool):
    description = Unicode(__doc__)
    name = 'ctapipe-dump-instrument'

    infile = Unicode(help='input simtelarray file').tag(config=True)
    format = Enum(['fits', 'ecsv', 'hdf5'],
                  default_value='fits',
                  help='Format of output file',
                  config=True)

    aliases = Dict(dict(infile='DumpInstrumentTool.infile',
                        format='DumpInstrumentTool.format'))


    def setup(self):
        source = hessio_event_source(self.infile)
        data = next(source)  # get one event, so the instrument table is there
        del source
        self.inst = data.inst  # keep a pointer to the instrument stuff
        pass

    def start(self):
        self.write_camera_geometries()
        self.write_optics_descriptions()
        self.write_subarray_description()

    def finish(self):
        pass

    def _get_file_format_info(self, format_name, table_type, table_name):
        """ returns file extension + dict of required parameters for 
        Table.write"""
        if format_name == 'fits':
            return 'fits.gz', dict()
        elif format_name == 'ecsv':
            return 'ecsv.txt', dict(format='ascii.ecsv')
        elif format_name == 'hdf5':
            return 'h5', dict(path="/" + table_type + "/" + table_name)
        else:
            raise NameError("format not supported")

    def write_camera_geometries(self):
        cam_types = get_camera_types(self.inst.subarray)
        self.inst.subarray.info(printer=self.log.info)
        for cam_name in cam_types:
            ext, args = self._get_file_format_info(self.format,
                                                   'CAMGEOM',
                                                   cam_name)

            self.log.debug("writing {}".format(cam_name))
            tel_id = cam_types[cam_name].pop()
            geom = self.inst.subarray.tel[tel_id].camera
            table = geom.to_table()
            table.meta['SOURCE'] = self.infile
            filename = "{}.camgeom.{}".format(cam_name, ext)

            try:
                table.write(filename, **args)
                Provenance().add_output_file(filename, 'dl0.tel.svc.camera')
            except IOError as err:
                self.log.warn("couldn't write camera definition '%s' because: "
                              "%s", filename, err)

    def write_optics_descriptions(self):
        sub = self.inst.subarray
        ext, args = self._get_file_format_info(self.format, sub.name, 'optics')

        tab = sub.to_table(kind='optics')
        tab.meta['SOURCE'] = self.infile
        filename = '{}.optics.{}'.format(sub.name, ext)
        try:
            tab.write(filename, **args)
            Provenance().add_output_file(filename, 'dl0.sub.svc.optics')
        except IOError as err:
            self.log.warn("couldn't write optics description '%s' because: "
                          "%s", filename, err)

    def write_subarray_description(self):
        sub = self.inst.subarray
        ext, args = self._get_file_format_info(self.format, sub.name,
                                               'subarray')
        tab = sub.to_table(kind='subarray')
        tab.meta['SOURCE'] = self.infile
        filename = '{}.subarray.{}'.format(sub.name, ext)
        try:
            tab.write(filename, **args)
            Provenance().add_output_file(filename, 'dl0.sub.svc.subarray')
        except IOError as err:
            self.log.warn("couldn't write subarray description '%s' because: "
                          "%s", filename, err)




def main():
    tool = DumpInstrumentTool()
    tool.run()

if __name__ == '__main__':
    main()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Visualization routines using matplotlib
"""
import copy
import logging

import matplotlib
import numpy as np
from astropy import units as u
from matplotlib import pyplot as plt
from matplotlib.collections import PatchCollection
from matplotlib.colors import Normalize, LogNorm, SymLogNorm
from matplotlib.lines import Line2D
from matplotlib.patches import Ellipse, RegularPolygon, Rectangle
from numpy import sqrt

__all__ = ['CameraDisplay', 'ArrayDisplay']

logger = logging.getLogger(__name__)

PIXEL_EPSILON = 0.0005 # a bit of extra size to pixels to avoid aliasing

class CameraDisplay:

    """
    Camera Display using matplotlib.

    Parameters
    ----------
    geometry : `~ctapipe.instrument.CameraGeometry`
        Definition of the Camera/Image
    image: array_like
        array of values corresponding to the pixels in the CameraGeometry.
    ax : `matplotlib.axes.Axes`
        A matplotlib axes object to plot on, or None to create a new one
    title : str (default "Camera")
        Title to put on camera plot
    norm : str or `matplotlib.color.Normalize` instance (default 'lin')
        Normalization for the color scale.
        Supported str arguments are
        - 'lin': linear scale
        - 'log': logarithmic scale (base 10)
    cmap : str or `matplotlib.colors.Colormap` (default 'hot')
        Color map to use (see `matplotlib.cm`)
    allow_pick : bool (default False)
        if True, allow user to click and select a pixel
    autoupdate : bool (default True)
        redraw automatically (otherwise need to call plt.draw())
    autoscale : bool (default True)
        rescale the vmin/vmax values when the image changes.
        This is set to False if `set_limits_*` is called to explicity
        set data limits.
    antialiased : bool  (default True)
        whether to draw in antialiased mode or not.

    Notes
    -----

    Speed:
        CameraDisplay is not intended to be very fast (matplotlib
        is not a very speed performant graphics library, it is
        intended for nice output plots). However, most of the
        slowness of CameraDisplay is in the constructor.  Once one is
        displayed, changing the image that is displayed is relatively
        fast and efficient. Therefore it is best to initialize an
        instance, and change the data, rather than generating new
        CameraDisplays.

    Pixel Implementation:
        Pixels are rendered as a
        `matplotlib.collections.PatchCollection` of Polygons (either 6
        or 4 sided).  You can access the PatchCollection directly (to
        e.g. change low-level style parameters) via
        `CameraDisplay.pixels`

    Output:
        Since CameraDisplay uses matplotlib, any display can be
        saved to any output file supported via
        plt.savefig(filename). This includes `.pdf` and `.png`.

    """

    def __init__(
            self,
            geometry,
            image=None,
            ax=None,
            title=None,
            norm="lin",
            cmap=None,
            allow_pick=False,
            autoupdate=True,
            autoscale=True,
            antialiased=True,
            ):
        self.axes = ax if ax is not None else plt.gca()
        self.geom = geometry
        self.pixels = None
        self.colorbar = None
        self.autoupdate = autoupdate
        self.autoscale = autoscale
        self._active_pixel = None
        self._active_pixel_label = None

        if title is None:
            title = geometry.cam_id

        # initialize the plot and generate the pixels as a
        # RegularPolyCollection

        patches = []

        if not hasattr(self.geom, "mask"):
            self.geom.mask = np.ones_like(self.geom.pix_x.value, dtype=bool)

        for xx, yy, aa in zip(
            u.Quantity(self.geom.pix_x[self.geom.mask]).value,
            u.Quantity(self.geom.pix_y[self.geom.mask]).value,
            u.Quantity(np.array(self.geom.pix_area)[self.geom.mask]).value):

            if self.geom.pix_type.startswith("hex"):
                rr = sqrt(aa * 2 / 3 / sqrt(3)) + 2*PIXEL_EPSILON
                poly = RegularPolygon(
                    (xx, yy), 6, radius=rr,
                    orientation=self.geom.pix_rotation.rad,
                    fill=True,
                )
            else:
                rr = sqrt(aa) + PIXEL_EPSILON
                poly = Rectangle(
                    (xx-rr/2., yy-rr/2.),
                    width=rr,
                    height=rr,
                    angle=self.geom.pix_rotation.deg,
                    fill=True,
                )

            patches.append(poly)

        self.pixels = PatchCollection(patches, cmap=cmap, linewidth=0)
        self.axes.add_collection(self.pixels)

        self.pixel_highlighting = copy.copy(self.pixels)
        self.pixel_highlighting.set_facecolor('none')
        self.pixel_highlighting.set_linewidth(0)
        self.axes.add_collection(self.pixel_highlighting)

        # Set up some nice plot defaults

        self.axes.set_aspect('equal', 'datalim')
        self.axes.set_title(title)
        self.axes.set_xlabel("X position ({})".format(self.geom.pix_x.unit))
        self.axes.set_ylabel("Y position ({})".format(self.geom.pix_y.unit))
        self.axes.autoscale_view()

        # set up a patch to display when a pixel is clicked (and
        # pixel_picker is enabled):

        self._active_pixel = copy.copy(patches[0])
        self._active_pixel.set_facecolor('r')
        self._active_pixel.set_alpha(0.5)
        self._active_pixel.set_linewidth(2.0)
        self._active_pixel.set_visible(False)
        self.axes.add_patch(self._active_pixel)

        self._active_pixel_label = self.axes.text(self._active_pixel.xy[0],
                                                  self._active_pixel.xy[1],
                                                  "0",
                                                  horizontalalignment='center',
                                                  verticalalignment='center')
        self._active_pixel_label.set_visible(False)

        # enable ability to click on pixel and do something (can be
        # enabled on-the-fly later as well:

        if allow_pick:
            self.enable_pixel_picker()

        if image is not None:
            self.image = image
        else:
            self.image = np.zeros_like(self.geom.pix_id, dtype=np.float)

        self.norm = norm

    def highlight_pixels(self, pixels, color='g', linewidth=1, alpha=0.75):
        '''
        Highlight the given pixels with a colored line around them

        Parameters
        ----------
        pixels : index-like
            The pixels to highlight.
            Can either be a list or array of integers or a
            boolean mask of length number of pixels
        color: a matplotlib conform color
            the color for the pixel highlighting
        linewidth: float
            linewidth of the highlighting in points
        alpha: 0 <= alpha <= 1
            The transparency
        '''

        l = np.zeros_like(self.image)
        l[pixels] = linewidth
        self.pixel_highlighting.set_linewidth(l)
        self.pixel_highlighting.set_alpha(alpha)
        self.pixel_highlighting.set_edgecolor(color)
        self._update()

    def enable_pixel_picker(self):
        """ enable ability to click on pixels """
        self.pixels.set_picker(True)  # enable click
        self.pixels.set_pickradius(sqrt(u.Quantity(self.geom.pix_area[0])
                                        .value) / np.pi)
        self.pixels.set_snap(True)  # snap cursor to pixel center
        self.axes.figure.canvas.mpl_connect('pick_event', self._on_pick)

    def set_limits_minmax(self, zmin, zmax):
        """ set the color scale limits from min to max """
        self.pixels.set_clim(zmin, zmax)
        self.autoscale = False
        self._update()

    def set_limits_percent(self, percent=95):
        """ auto-scale the color range to percent of maximum """
        zmin = self.pixels.get_array().min()
        zmax = self.pixels.get_array().max()
        dz = zmax - zmin
        frac = percent / 100.0
        self.autoscale = False
        self.set_limits_minmax(zmin, zmax - (1.0 - frac) * dz)

    @property
    def norm(self):
        '''
        The norm instance of the Display

        Possible values:

        - "lin": linear scale
        - "log": log scale (cannot have negative values)
        - "symlog": symmetric log scale (negative values are ok)
        -  any matplotlib.colors.Normalize instance, e. g. PowerNorm(gamma=-2)
        '''
        return self.pixels.norm

    @norm.setter
    def norm(self, norm):

        if norm == 'lin':
            self.pixels.norm = Normalize()
        elif norm == 'log':
            self.pixels.norm = LogNorm()
            self.pixels.autoscale()  # this is to handle matplotlib bug #5424
        elif norm == 'symlog':
            self.pixels.norm = SymLogNorm(linthresh=1.0)
            self.pixels.autoscale()
        elif isinstance(norm, Normalize):
            self.pixels.norm = norm
        else:
            raise ValueError("Unsupported norm: '{}', options are 'lin',"
                             "'log','symlog', or a matplotlib Normalize object"
                             .format(norm))

        self.update(force=True)
        self.pixels.autoscale()

    @property
    def cmap(self):
        """
        Color map to use. Either a name or  `matplotlib.colors.ColorMap`
        instance, e.g. from `matplotlib.pyplot.cm`
        """
        return self.pixels.get_cmap()

    @cmap.setter
    def cmap(self, cmap):
        self.pixels.set_cmap(cmap)
        self._update()

    @property
    def image(self):
        """The image displayed on the camera (1D array of pixel values)"""
        return self.pixels.get_array()

    @image.setter
    def image(self, image):
        """
        Change the image displayed on the Camera.

        Parameters
        ----------
        image: array_like
            array of values corresponding to the pixels in the CameraGeometry.
        """
        image = np.asanyarray(image)
        if image.shape != self.geom.pix_x.shape:
            raise ValueError(
                "Image has a different shape {} than the "
                "given CameraGeometry {}"
                .format(image.shape, self.geom.pix_x.shape)
            )

        self.pixels.set_array(image[self.geom.mask])
        self.pixels.changed()
        if self.autoscale:
            self.pixels.autoscale()
        self._update()

    def _update(self, force=False):
        """ signal a redraw if autoupdate is turned on """
        if self.autoupdate:
            self.update(force)

    def update(self, force=False):
        """ redraw the display now """
        self.axes.figure.canvas.draw()
        if self.colorbar is not None:
            if force is True:
                self.colorbar.update_bruteforce(self.pixels)
            else:
                self.colorbar.update_normal(self.pixels)
            self.colorbar.draw_all()

    def add_colorbar(self, **kwargs):
        """
        add a colobar to the camera plot
        kwargs are passed to `figure.colorbar(self.pixels, **kwargs)`
        See matplotlib documentation for the supported kwargs:
        http://matplotlib.org/api/figure_api.html#matplotlib.figure.Figure.colorbar
        """
        if self.colorbar is not None:
            raise ValueError(
                'There is already a colorbar attached to this CameraDisplay'
            )
        else:
            self.colorbar = self.axes.figure.colorbar(self.pixels, **kwargs)
        self.update()

    def add_ellipse(self, centroid, length, width, angle, asymmetry=0.0,
                    **kwargs):
        """
        plot an ellipse on top of the camera

        Parameters
        ----------
        centroid: (float, float)
            position of centroid
        length: float
            major axis
        width: float
            minor axis
        angle: float
            rotation angle wrt x-axis about the centroid, anticlockwise, in radians
        asymmetry: float
            3rd-order moment for directionality if known
        kwargs:
            any MatPlotLib style arguments to pass to the Ellipse patch

        """
        ellipse = Ellipse(xy=centroid, width=length, height=width,
                          angle=np.degrees(angle), fill=False, **kwargs)
        self.axes.add_patch(ellipse)
        self.update()
        return ellipse

    def overlay_moments(self, momparams, with_label=True, **kwargs):
        """helper to overlay ellipse from a `reco.MomentParameters` structure

        Parameters
        ----------
        momparams: `reco.MomentParameters`
            structuring containing Hillas-style parameterization
        kwargs: key=value
            any style keywords to pass to matplotlib (e.g. color='red'
            or linewidth=6)
        """

        # strip off any units
        cen_x = u.Quantity(momparams.cen_x).value
        cen_y = u.Quantity(momparams.cen_y).value
        length = u.Quantity(momparams.length).value
        width = u.Quantity(momparams.width).value


        el = self.add_ellipse(centroid=(cen_x, cen_y),
                              length=length*2,
                              width=width*2, angle=momparams.psi.rad,
                              **kwargs)
        if with_label:
            self.axes.text(cen_x, cen_y,
                           ("({:.02f},{:.02f})\n"
                            "[w={:.02f},l={:.02f}]")
                           .format(momparams.cen_x,
                                   momparams.cen_y,
                                   momparams.width, momparams.length),
                           color=el.get_edgecolor())

    def _on_pick(self, event):
        """ handler for when a pixel is clicked """
        pix_id = event.ind[-1]
        xx, yy, aa = u.Quantity(self.geom.pix_x[pix_id]).value, \
                     u.Quantity(self.geom.pix_y[pix_id]).value, \
                     u.Quantity(np.array(self.geom.pix_area)[pix_id])
        if self.geom.pix_type.startswith("hex"):
            self._active_pixel.xy = (xx, yy)
        else:
            rr = sqrt(aa)
            self._active_pixel.xy = (xx - rr / 2., yy - rr / 2.)
        self._active_pixel.set_visible(True)
        self._active_pixel_label.set_x(xx)
        self._active_pixel_label.set_y(yy)
        self._active_pixel_label.set_text("{:003d}".format(pix_id))
        self._active_pixel_label.set_visible(True)
        self._update()
        self.on_pixel_clicked(pix_id)  # call user-function

    def on_pixel_clicked(self, pix_id):
        """virtual function to overide in sub-classes to do something special
        when a pixel is clicked
        """
        print("Clicked pixel_id {}".format(pix_id))

    def show(self):
        self.axes.figure.show()


class ArrayDisplay:

    """
    Display a top-town view of a telescope array
    """

    def __init__(self, telx, tely, tel_type=None, radius=20,
                 axes=None, title="Array", autoupdate=True):

        if tel_type is None:
            tel_type = np.ones(len(telx))
        patches = [Rectangle(xy=(x-radius/2, y-radius/2), width=radius, height=radius, fill=False)
                   for x, y in zip(telx, tely)]

        self.autoupdate = autoupdate
        self.telescopes = PatchCollection(patches, match_original=True)
        self.telescopes.set_clim(1, 9)
        rgb = matplotlib.cm.Set1((tel_type-1)/9)
        self.telescopes.set_edgecolor(rgb)
        self.telescopes.set_linewidth(2.0)

        self.axes = axes if axes is not None else plt.gca()
        self.axes.add_collection(self.telescopes)
        self.axes.set_aspect(1.0)
        self.axes.set_title(title)
        self.axes.set_xlim(-1000, 1000)
        self.axes.set_ylim(-1000, 1000)

        self.axes_hillas = axes if axes is not None else plt.gca()


    @property
    def values(self):
        """An array containing a value per telescope"""
        return self.telescopes.get_array()

    @values.setter
    def values(self, values):
        """ set the telescope colors to display  """
        self.telescopes.set_array(values)
        self._update()

    def _update(self):
        """ signal a redraw if necessary """
        if self.autoupdate:
            plt.draw()

    def add_ellipse(self, centroid, length, width, angle, **kwargs):
        """
        plot an ellipse on top of the camera

        Parameters
        ----------
        centroid: (float, float)
            position of centroid
        length: float
            major axis
        width: float
            minor axis
        angle: float
            rotation angle wrt x-axis about the centroid, anticlockwise, in radians
        asymmetry: float
            3rd-order moment for directionality if known
        kwargs:
            any MatPlotLib style arguments to pass to the Ellipse patch

        """
        ellipse = Ellipse(xy=centroid, width=length, height=width,
                          angle=np.degrees(angle), fill=True,  **kwargs)
        self.axes.add_patch(ellipse)
        return ellipse

    def add_polygon(self, centroid, radius, nsides=3, **kwargs):
        """
        plot a polygon on top of the camera

        Parameters
        ----------
        centroid: (float, float)
            position of centroid
        radius: float
            radius
        nsides: int
            Number of points on polygon
        kwargs:
            any MatPlotLib style arguments to pass to the RegularPolygon patch

        """
        polygon = RegularPolygon(xy=centroid, radius=radius, numVertices=nsides, **kwargs)
        self.axes.add_patch(polygon)
        return polygon

    def overlay_moments(self, momparams, tel_position, scale_fac, **kwargs):
        """helper to overlay ellipse from a `reco.MomentParameters` structure

        Parameters
        ----------
        momparams: `reco.MomentParameters`
            structuring containing Hillas-style parameterization
        tel_position: list
            (x, y) positions of each telescope
        scale_fac: float
            scaling factor to apply to width and length when overlaying moments
        kwargs: key=value
            any style keywords to pass to matplotlib (e.g. color='red'
            or linewidth=6)
        """
        # strip off any units
        ellipse_list = list()
        size_list = list()
        i = 0
        for h in momparams:

            length = u.Quantity(momparams[h].length).value * scale_fac
            width = u.Quantity(momparams[h].width).value * scale_fac
            size_list.append(u.Quantity(momparams[h].size).value)
            tel_x = u.Quantity(tel_position[0][i]).value
            tel_y = u.Quantity(tel_position[1][i]).value
            i += 1

            ellipse = Ellipse(xy=(tel_x,tel_y), width=length, height=width,
                              angle=np.degrees(momparams[h].psi.rad))
            ellipse_list.append(ellipse)

        patches = PatchCollection(ellipse_list, **kwargs)
        patches.set_clim(0, 1000) # Set ellipse colour based on image size
        patches.set_array(np.asarray(size_list))
        self.axes_hillas.add_collection(patches)

    def overlay_axis(self, momparams, tel_position, **kwargs):
        """helper to overlay ellipse from a `reco.MomentParameters` structure

        Parameters
        ----------
        momparams: `reco.MomentParameters`
            structuring containing Hillas-style parameterization
        tel_position: list
            (x, y) positions of each telescope
        kwargs: key=value
            any style keywords to pass to matplotlib (e.g. color='red'
            or linewidth=6)
        """
        # strip off any units
        i = 0
        for h in momparams:
            tel_x = u.Quantity(tel_position[0][i]).value
            tel_y = u.Quantity(tel_position[1][i]).value
            psi = u.Quantity(momparams[h].psi).value
            x_sc = [tel_x - np.cos(psi) * 10000, tel_x + np.cos(psi) * 10000]
            y_sc = [tel_y - np.sin(psi) * 10000, tel_y + np.sin(psi) * 10000]

            i += 1
            self.axes_hillas.add_line(Line2D(x_sc, y_sc, linestyle='dashed', color='black'))



'''A simple example of how to use traitlets.config.application.Application.

This should serve as a simple example that shows how the traitlets config
system works. The main classes are:

* traitlets.config.Configurable
* traitlets.config.SingletonConfigurable
* traitlets.config.Config
* traitlets.config.Application

To see the command line option help, run this program from the command line::

    $ python myapp.py -h

To make one of your classes configurable (from the command line and config
files) inherit from Configurable and declare class attributes as traits (see
classes Foo and Bar below). To make the traits configurable, you will need
to set the following options:

* ``config``: set to ``True`` to make the attribute configurable.
* ``shortname``: by default, configurable attributes are set using the syntax
  'Classname.attributename'. At the command line, this is a bit verbose, so
  we allow 'shortnames' to be declared. Setting a shortname is optional, but
  when you do this, you can set the option at the command line using the
  syntax: 'shortname=value'.
* ``help``: set the help string to display a help message when the ``-h``
  option is given at the command line. The help string should be valid ReST.

When the config attribute of an Application is updated, it will fire all of
the trait's events for all of the config=True attributes.
'''

from traitlets.config.configurable import Configurable
from traitlets.config.application import Application
from traitlets import (
    Bool, Unicode, Int, List, Dict
)

from ctapipe.utils.json2fits import traitlets_config_to_fits, json_to_fits
from ctapipe.utils import get_dataset
import tempfile

import sys
import os
import json


class Foo(Configurable):
    '''A class that has configurable, typed attributes.

    '''

    i = Int(0, help='The integer i.').tag(config=True)
    j = Int(1, help='The integer j.').tag(config=True)
    name = Unicode(u'Brian', help='First name.').tag(config=True)


class Bar(Configurable):

    enabled = Bool(True, help='Enable bar.').tag(config=True)


class MyApp(Application):

    name = Unicode('myapp')
    running = Bool(False,
                   help='Is the app running?').tag(config=True)
    classes = List([Bar, Foo])

    aliases = Dict(dict(
        i='Foo.i', j='Foo.j', name='Foo.name', running='MyApp.running',
        enabled='Bar.enabled', log_level='MyApp.log_level',
        config_file='MyApp.config_file',
    ))

    flags = Dict(dict(
        enable=({'Bar': {'enabled': True}}, 'Enable Bar'),
        disable=({'Bar': {'enabled': False}}, 'Disable Bar'),
        debug=({'MyApp': {'log_level': 10}}, 'Set loglevel to DEBUG'),

    ))
    key_much_too_long_for_fits = Unicode('value short enough')
    key_and_value_too_long = Unicode(5 * 'value to long as wellenough')
    short = Unicode(20 * 'value to long')
    config_file = Unicode(u'', help='Load this config file').tag(config=True)

    def init_foo(self):
        # Pass config to other classes for them to inherit the config.
        self.foo = Foo(config=self.config)

    def init_bar(self):
        # Pass config to other classes for them to inherit the config.
        self.bar = Bar(config=self.config)

    def initialize(self, argv=None):
        self.parse_command_line(argv)
        self.full_path_configfile = self.config_file
        if self.config_file:
            self.load_config_file(self.config_file)
        self.init_foo()
        self.init_bar()

    def start(self):
        pass

    def stage_default_config_file(self):
        '''auto generate default config file, and stage it into the profile.'''
        # s = self.generate_config_file()
        fname = os.path.join('.', 'foo.json')  # self.config_file)
        if not os.path.exists(fname):
            self.log.warn('Generating default config file: {}'.format(fname))
            with open(fname, 'w') as f:
                f.write(str(json.dumps(self.config)))

    def traitlets_config_to_fits(self, outputfile):
        traitlets_config_to_fits(self.config, outputfile, clobber=True)

    def jsonToFits(self, outputfile):
        json_to_fits(self.full_path_configfile, outputfile, clobber=True)


def test_traitlets_config_to_fits():
    backup = sys.argv
    full_config_name = get_dataset('config.json')
    sys.argv = ['test_json_2_fits.py', '--config_file=' + full_config_name]
    app = MyApp()
    app.initialize()
    app.start()

    tmp = tempfile.NamedTemporaryFile()
    app.traitlets_config_to_fits(tmp.name)
    sys.argv = backup


def test_jsonToFits():
    backup = sys.argv
    full_config_name = get_dataset('config.json')
    sys.argv = ['test_json_2_fits.py', '--config_file=' + full_config_name]
    app = MyApp()
    app.initialize()
    app.start()
    tmp = tempfile.NamedTemporaryFile()
    app.jsonToFits(tmp)
    sys.argv = backup


def main():
    test_traitlets_config_to_fits()
    test_jsonToFits()


if __name__ == '__main__':
    main()



""" test code to ensure pedestal functions behave correctly """
from .. import pedestals
import numpy as np


def test_calc_pedestals_from_traces():

    # create some test data (all ones, but with a 2 stuck in for good measure):
    npix = 1000
    nsamp = 32
    traces = np.ones(shape=(npix, nsamp), dtype=np.int32)
    traces[:, 2] = 2

    # calculate pedestals over samples 0-10
    peds, pedvars = pedestals.calc_pedestals_from_traces(traces, 0, 10)

    # check the results. All pedestals should be 1.1 and vars 0.09:

    assert np.all(peds == 1.1)
    assert np.all(pedvars == 0.09)
    assert peds.shape[0] == npix
    assert pedvars.shape[0] == npix

    # try another sample range, where there should be no variances and
    # all 1.0 peds:
    peds, pedvars = pedestals.calc_pedestals_from_traces(traces, 12, 32)

    assert np.all(peds == 1.0)
    assert np.all(pedvars == 0)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Create a toymodel event stream of array events
"""
import logging

import numpy as np
from ctapipe.image import toymodel
from scipy.stats import norm

from .containers import DataContainer, R0CameraContainer

logger = logging.getLogger(__name__)


def toymodel_event_source(geoms, max_events=100, single_tel=False, n_channels=1,
                          n_samples=25, p_trigger=0.3):
    """
    An event source that produces array
    Parameters
    ----------
    geoms : list of CameraGeometry instances
        Geometries for the telescopes to simulate
    max_events : int, default: 100
        maximum number of events to create
    n_channels : int
        how many channels per telescope
    n_samples : int
        how many adc samples per pixel
    p_trigger : float
        mean trigger probability for the telescopes
    """
    n_telescopes = len(geoms)
    container = DataContainer()
    container.meta['toymodel__max_events'] = max_events
    container.meta['source'] = "toymodel"
    tel_ids = np.arange(n_telescopes)

    for event_id in range(max_events):

        n_triggered = np.random.poisson(n_telescopes * 0.3)
        if n_triggered > n_telescopes:
            n_triggered = n_telescopes

        triggered_tels = np.random.choice(tel_ids, n_triggered, replace=False)

        container.r0.event_id = event_id
        container.r0.tels_with_data = triggered_tels
        container.count = event_id

        # handle single-telescope case (ignore others:
        if single_tel:
            if single_tel not in container.r0.tels_with_data:
                continue
            container.r0.tels_with_data = [single_tel, ]

        container.r0.tel.reset()  # clear the previous telescopes
        t = np.arange(n_samples)

        for tel_id in container.r0.tels_with_data:
            geom = geoms[tel_id]

            # fill pixel position dictionary, if not already done:
            if tel_id not in container.inst.pixel_pos:
                container.inst.pixel_pos[tel_id] = (
                    geom.pix_x.value,
                    geom.pix_y.value,
                )

            centroid = np.random.uniform(geom.pix_x.min(), geom.pix_y.max(), 2)
            length = np.random.uniform(0.02, 0.2)
            width = np.random.uniform(0.01, length)
            psi = np.random.randint(0, 360)
            intensity = np.random.poisson(int(10000 * width * length))
            model = toymodel.generate_2d_shower_model(
                centroid,
                width,
                length,
                '{}d'.format(psi)
            )
            image, _, _ = toymodel.make_toymodel_shower_image(
                geom,
                model.pdf,
                intensity,
            )

            # container.r0.tel[tel_id] = R0CameraContainer()
            container.inst.num_channels[tel_id] = n_channels
            n_pix = len(geom.pix_id)
            samples = np.empty((n_pix, n_samples))
            means = np.random.normal(15, 1, (n_pix, 1))
            stds = np.random.uniform(3, 6, (n_pix, 1))
            samples = image[:, np.newaxis] * norm.pdf(t, means, stds)

            for chan in range(n_channels):
                container.r0.tel[tel_id].adc_samples[chan] = samples
                container.r0.tel[tel_id].adc_sums[chan] = image

        yield container



import numpy as np
import pytest
from astropy import units as u
from astropy.coordinates import SkyCoord, AltAz

from ctapipe.calib.camera.dl0 import CameraDL0Reducer
from ctapipe.calib.camera.dl1 import CameraDL1Calibrator
from ctapipe.calib.camera.r1 import HessioR1Calibrator
from ctapipe.coordinates import NominalFrame, CameraFrame
from ctapipe.image.cleaning import tailcuts_clean
from ctapipe.image.hillas import hillas_parameters, HillasParameterizationError
from ctapipe.io.hessio import hessio_event_source
from ctapipe.plotting.array import NominalPlotter
from ctapipe.utils import get_dataset


@pytest.mark.skip
def test_array_draw():
    filename = get_dataset("gamma_test.simtel.gz")
    cam_geom = {}

    source = hessio_event_source(filename, max_events=2)
    r1 = HessioR1Calibrator(None, None)
    dl0 = CameraDL0Reducer(None, None)

    calibrator = CameraDL1Calibrator(None, None)

    for event in source:
        array_pointing = SkyCoord(event.mcheader.run_array_direction[1] * u.rad,
                                  event.mcheader.run_array_direction[0] * u.rad,
                                  frame=AltAz)
        # array_view = ArrayPlotter(instrument=event.inst,
        #                          system=TiltedGroundFrame(
        # pointing_direction=array_pointing))

        hillas_dict = {}
        r1.calibrate(event)
        dl0.reduce(event)
        calibrator.calibrate(event)  # calibrate the events

        # store MC pointing direction for the array

        for tel_id in event.dl0.tels_with_data:

            pmt_signal = event.dl1.tel[tel_id].image[0]
            geom = event.inst.subarray.tel[tel_id].camera
            fl = event.inst.subarray.tel[tel_id].optics.effective_focal_length

            # Transform the pixels positions into nominal coordinates
            camera_coord = CameraFrame(x=geom.pix_x, y=geom.pix_y,
                                       z=np.zeros(geom.pix_x.shape) * u.m,
                                       focal_length=fl,
                                       rotation=90 * u.deg - geom.cam_rotation)

            nom_coord = camera_coord.transform_to(
                NominalFrame(array_direction=array_pointing,
                             pointing_direction=array_pointing))

            mask = tailcuts_clean(geom, pmt_signal,
                                  picture_thresh=10., boundary_thresh=5.)

            try:
                moments = hillas_parameters(nom_coord.x,
                                            nom_coord.y,
                                            pmt_signal * mask)
                hillas_dict[tel_id] = moments
                nom_coord = NominalPlotter(hillas_parameters=hillas_dict,
                                           draw_axes=True)
                nom_coord.draw_array()

            except HillasParameterizationError as e:
                print(e)
                continue

                # array_view.background_image(np.ones((4,4)), ((-1500,1500),
                # (-1500,1500)))
                # array_view.overlay_hillas(hillas_dict, draw_axes=True)
                # array_view.draw_array(range=((-1000,1000),(-1000,1000)))
                # return


if __name__ == "__main__":
    test_array_draw()




"""
Set of diagnostic plots relating to muons 
For generic use with all muon algorithms
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.mlab as mlab
from matplotlib import colors
from scipy.stats import norm
from astropy import units as u
from astropy.table import Table
from ctapipe.instrument import CameraGeometry
from ctapipe.coordinates import (CameraFrame, NominalFrame, HorizonFrame,
                                 TelescopeFrame)
from ctapipe.image.cleaning import tailcuts_clean
from IPython import embed

from ctapipe.plotting.camera import CameraPlotter
from ctapipe.utils.fitshistogram import Histogram


def plot_muon_efficiency(outputpath):
    """
    Plot the muon efficiencies
    """
    fig, ax = plt.subplots(1, 1, figsize=(10, 10))
    figip, axip = plt.subplots(1, 1, figsize=(10, 10))
    figrw, axrw = plt.subplots(1, 1, figsize=(10, 10))

    nbins = 16
    t = Table.read(str(outputpath) + '_muontable.fits')
    print('Reading muon efficiency from table', outputpath, t['MuonEff'])

    if len(t['MuonEff']) < 1:
        print("No muon events to plot")
        return
    else:
        print("Found", len(t['MuonEff']), "muon events")

    (mu, sigma) = norm.fit(t['MuonEff'])

    print('Gaussian fit with mu=', mu, 'sigma=', sigma)

    conteff = ax.hist(t['MuonEff'], nbins)
    ax.set_xlim(0.2 * min(t['MuonEff']), 1.2 * max(t['MuonEff']))

    xtest = np.linspace(min(t['MuonEff']), max(t['MuonEff']), nbins)
    yg = mlab.normpdf(xtest, mu, sigma)
    print('mu', mu, 'sigma', sigma, 'yg', yg)
    ax.plot(xtest, yg, 'r', linewidth=2)

    ax.set_ylim(0., 1.2 * max(conteff[0]))
    ax.set_xlabel('Muon Efficiency')
    plt.draw()

    yimp = np.linspace(min(t['ImpactP']), max(t['ImpactP']), nbins)
    contimp = axip.hist(t['ImpactP'], nbins)
    axip.set_xlim(0.2 * min(t['ImpactP']), 1.2 * max(t['ImpactP']))
    axip.set_ylim(0., 1.2 * max(contimp[0]))
    axip.set_xlabel('Impact Parameter (m)')

    plt.draw()

    heffimp = Histogram(nbins=[16, 16],
                        ranges=[(min(t['MuonEff']), max(t['MuonEff'])),
                                (min(t['ImpactP']), max(t['ImpactP']))])  # ,axisNames=["MuonEfficiency","ImpactParameter"])

    heffimp.fill([t['MuonEff'], t['ImpactP']])
    heffimp.draw_2d()

    yrw = np.linspace(min(t['RingWidth']), max(t['RingWidth']), nbins)
    contrw = axrw.hist(t['RingWidth'], nbins)
    axrw.set_xlim(0.2 * min(t['RingWidth']), 1.2 * max(t['RingWidth']))
    axrw.set_ylim(0., 1.2 * max(contrw[0]))
    axrw.set_xlabel('Ring Width ($^\circ$)')

    plt.draw()

    if outputpath is not None:
        print("saving figure at", outputpath)
        fig.savefig(str(outputpath) + '_MuonEff.png')
        figip.savefig(str(outputpath) + '_ImpactParameter.png')
        figrw.savefig(str(outputpath) + '_RingWidth.png')
    else:
        print("Not saving figure, no outputpath")
        plt.show()


def plot_muon_event(event, muonparams, args=None):

    if muonparams['MuonRingParams'] is not None:

        # Plot the muon event and overlay muon parameters
        fig = plt.figure(figsize=(16, 7))

        colorbar = None
        colorbar2 = None

        #for tel_id in event.dl0.tels_with_data:
        for tel_id in muonparams['TelIds']:
            idx = muonparams['TelIds'].index(tel_id)

            if not muonparams['MuonRingParams'][idx]:
                continue

            #otherwise...
            npads = 2
            # Only create two pads if there is timing information extracted
            # from the calibration
            ax1 = fig.add_subplot(1, npads, 1)
            plotter = CameraPlotter(event)
            image = event.dl1.tel[tel_id].image[0]
            geom = event.inst.subarray.tel[tel_id].camera

            tailcuts = (5., 7.)
            # Try a higher threshold for
            if geom.cam_id == 'FlashCam':
                tailcuts = (10., 12.)

            clean_mask = tailcuts_clean(geom, image,
                                        picture_thresh=tailcuts[0],
                                        boundary_thresh=tailcuts[1])

            signals = image * clean_mask

            #print("Ring Centre in Nominal Coords:",muonparams[0].ring_center_x,muonparams[0].ring_center_y)
            muon_incl = np.sqrt(muonparams['MuonRingParams'][idx].ring_center_x**2. +
                                muonparams['MuonRingParams'][idx].ring_center_y**2.)

            muon_phi = np.arctan(muonparams['MuonRingParams'][idx].ring_center_y /
                                 muonparams['MuonRingParams'][idx].ring_center_x)

            rotr_angle = geom.pix_rotation
            # if event.inst.optical_foclen[tel_id] > 10.*u.m and
            # event.dl0.tel[tel_id].num_pixels != 1764:
            if geom.cam_id == 'LSTCam' or geom.cam_id == 'NectarCam':
                #print("Resetting the rotation angle")
                rotr_angle = 0. * u.deg

            # Convert to camera frame (centre & radius)
            altaz = HorizonFrame(alt=event.mc.alt, az=event.mc.az)

            ring_nominal = NominalFrame(x=muonparams['MuonRingParams'][idx].ring_center_x,
                                        y=muonparams['MuonRingParams'][idx].ring_center_y,
                                        array_direction=altaz,
                                        pointing_direction=altaz)

            # embed()
            ring_camcoord = ring_nominal.transform_to(CameraFrame(
                pointing_direction=altaz,
                focal_length=event.inst.optical_foclen[tel_id],
                rotation=rotr_angle))

            centroid_rad = np.sqrt(ring_camcoord.y**2 + ring_camcoord.x**2)
            centroid = (ring_camcoord.x.value, ring_camcoord.y.value)

            ringrad_camcoord = muonparams['MuonRingParams'][idx].ring_radius.to(u.rad) \
                               * event.inst.optical_foclen[tel_id] * 2.  # But not FC?

            px, py = event.inst.pixel_pos[tel_id]
            flen = event.inst.optical_foclen[tel_id]
            camera_coord = CameraFrame(x=px, y=py,
                                       z=np.zeros(px.shape) * u.m,
                                       focal_length=flen,
                                       rotation=geom.pix_rotation)

            nom_coord = camera_coord.transform_to(
                NominalFrame(array_direction=altaz,
                             pointing_direction=altaz)
            )

            px = nom_coord.x.to(u.deg)
            py = nom_coord.y.to(u.deg)

            dist = np.sqrt(np.power( px - muonparams['MuonRingParams'][idx].ring_center_x, 2)
                           + np.power(py - muonparams['MuonRingParams'][idx].ring_center_y, 2))
            ring_dist = np.abs(dist - muonparams['MuonRingParams'][idx].ring_radius)
            pixRmask = ring_dist < muonparams['MuonRingParams'][idx].ring_radius * 0.4

            #if muonparams[1] is not None:
            if muonparams['MuonIntensityParams'][idx] is not None:
                signals *= muonparams['MuonIntensityParams'][idx].mask

            camera1 = plotter.draw_camera(tel_id, signals, ax1)

            cmaxmin = (max(signals) - min(signals))
            cmin = min(signals)
            if not cmin:
                cmin = 1.
            if not cmaxmin:
                cmaxmin = 1.

            cmap_charge = colors.LinearSegmentedColormap.from_list(
                'cmap_c', [(0 / cmaxmin, 'darkblue'),
                           (np.abs(cmin) / cmaxmin, 'black'),
                           (2.0 * np.abs(cmin) / cmaxmin, 'blue'),
                           (2.5 * np.abs(cmin) / cmaxmin, 'green'),
                           (1, 'yellow')]
            )
            camera1.pixels.set_cmap(cmap_charge)
            if not colorbar:
                camera1.add_colorbar(ax=ax1, label=" [photo-electrons]")
                colorbar = camera1.colorbar
            else:
                camera1.colorbar = colorbar
            camera1.update(True)

            camera1.add_ellipse(centroid, ringrad_camcoord.value,
                                ringrad_camcoord.value, 0., 0., color="red")


            if muonparams['MuonIntensityParams'][idx] is not None:
                # continue #Comment this...(should ringwidthfrac also be *0.5?)

                ringwidthfrac = muonparams['MuonIntensityParams'][idx].ring_width / muonparams['MuonRingParams'][idx].ring_radius
                ringrad_inner = ringrad_camcoord * (1. - ringwidthfrac)
                ringrad_outer = ringrad_camcoord * (1. + ringwidthfrac)
                camera1.add_ellipse(centroid, ringrad_inner.value,
                                     ringrad_inner.value, 0., 0.,
                                     color="magenta")
                camera1.add_ellipse(centroid, ringrad_outer.value,
                                    ringrad_outer.value, 0., 0., color="magenta")
                npads = 2
                ax2 = fig.add_subplot(1, npads, npads)
                pred = muonparams['MuonIntensityParams'][idx].prediction


                if len(pred) != np.sum(muonparams['MuonIntensityParams'][idx].mask):
                    print("Warning! Lengths do not match...len(pred)=",
                          len(pred), "len(mask)=", np.sum(muonparams['MuonIntensityParams'][idx].mask))

                # Numpy broadcasting - fill in the shape
                plotpred = np.zeros(image.shape)
                plotpred[muonparams['MuonIntensityParams'][idx].mask == True] = pred

                camera2 = plotter.draw_camera(tel_id, plotpred, ax2)

                if np.isnan(max(plotpred)) or np.isnan(min(plotpred)):
                    print("nan prediction, skipping...")
                    continue

                c2maxmin = (max(plotpred) - min(plotpred))
                if not c2maxmin:
                    c2maxmin = 1.

                c2map_charge = colors.LinearSegmentedColormap.from_list(
                    'c2map_c', [(0 / c2maxmin, 'darkblue'),
                                (np.abs(min(plotpred)) / c2maxmin, 'black'),
                                (2.0 * np.abs(min(plotpred)) / c2maxmin, 'blue'),
                                (2.5 * np.abs(min(plotpred)) / c2maxmin, 'green'),
                                (1, 'yellow')]
                )
                camera2.pixels.set_cmap(c2map_charge)
                if not colorbar2:
                    camera2.add_colorbar(ax=ax2, label=" [photo-electrons]")
                    colorbar2 = camera2.colorbar
                else:
                    camera2.colorbar = colorbar2
                camera2.update(True)
                plt.pause(1.)  # make shorter


            # plt.pause(0.1)
            # if pp is not None:
            #    pp.savefig(fig)
            #fig.savefig(str(args.output_path) + "_" +
            #            str(event.dl0.event_id) + '.png')


            plt.close()



from ctapipe.core.factory import Factory
from ctapipe.core.component import Component
from traitlets import Unicode, Int


class ExampleComponentParent(Component):
    name = 'ExampleComponentParent'
    value = Int(123, help="").tag(config=True)


class ExampleComponent1(ExampleComponentParent):
    name = 'ExampleComponent1'
    value = Int(123111, help="").tag(config=True)


class ExampleComponent2(ExampleComponentParent):
    name = 'ExampleComponent2'
    value = Int(123222, help="").tag(config=True)


class ExampleFactory(Factory):
    name = 'ExampleFactory'
    description = "Test Factory class"

    subclasses = Factory.child_subclasses(ExampleComponentParent)
    subclass_names = [c.__name__ for c in subclasses]

    discriminator = Unicode('ExampleComponent1',
                            help='Product to obtain: {}'
                            .format(subclass_names)).tag(config=True)

    # Product classes traits
    value = Int(555, help="").tag(config=True)

    def get_factory_name(self):
        return self.name

    def get_product_name(self):
        return self.discriminator


def test_factory():
    factory = ExampleFactory(config=None, tool=None)
    factory.discriminator = 'ExampleComponent2'
    factory.value = 111
    cls = factory.get_class()
    obj = cls(config=factory.config, parent=None)
    assert(obj.name == 'ExampleComponent2')
    assert(obj.value == 111)



import numpy as np

from astropy import units as u

from sklearn.ensemble import RandomForestClassifier

from .regressor_classifier_base import RegressorClassifierBase


def proba_drifting(x):
    """
    gives more weight to outliers -- i.e. close to 0 and 1
    the polynomial was constructed with the following constraints:
    • f(0) = 0
    • f(0.5) = 0.5
    • f(1) = 1
    • f'(0) = 0
    • f'(0.5) = 1
    • f'(1) = 0
    """
    return 10 * x**3 - 15 * x**4 + 6 * x**5


class EventClassifier(RegressorClassifierBase):
    def __init__(self, classifier=RandomForestClassifier,
                 cam_id_list=("cam"), **kwargs):
        super().__init__(model=classifier, cam_id_list=cam_id_list, **kwargs)

    def predict_proba_by_event(self, X):
        predict_proba = []
        for evt in X:
            tel_probas = None
            tel_weights = []
            for cam_id, tels in evt.items():
                these_probas = self.model_dict[cam_id].predict_proba(tels)
                tel_probas = np.append(these_probas, tel_probas, axis=0) \
                    if tel_probas is not None else these_probas
                try:
                    # if a `namedtuple` is provided, we can weight the different images
                    # using some of the provided features
                    tel_weights += [t.sum_signal_cam / t.impact_dist for t in tels]
                except AttributeError:
                    # otherwise give every image the same weight
                    tel_weights += [1] * len(tels)

            predict_proba.append(np.average(proba_drifting(tel_probas),
                                            weights=tel_weights, axis=0))

        return np.array(predict_proba)

    def predict_by_event(self, X):
        proba = self.predict_proba_by_event(X)
        predictions = self.classes_[np.argmax(proba, axis=1)]
        return predictions



from ctapipe.core import Component
from time import sleep

class IntGenerator(Component):

    """IntGenerator class represents a Producer for pipeline.
    It yields integer from 0 to 5 to Pair or Odd stages, depending of their parity
    """
    def init(self):
        return True

    def run(self):
        for i in range(15):
            sleep(.5)
            self.log.debug("IntGenerator send {}".format(i))
            if i % 2 == 0:
                yield (i,'Pair')
            else: yield(i,'Odd')

        self.log.debug("\n--- IntGenerator Done ---")

    def finish(self):
        self.log.debug("--- IntGenerator finish ---")
        pass
